#include "fracram.h"

/***********************************************************/
/* GUI ROUTINES                                            */
/***********************************************************/
/****************************************/
/* Main window display.                 */
/****************************************/
void DISPLAY_MAIN_PANEL(void) {
   glDrawBuffer(GL_BACK);
   glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
   glColor3f(0.6, 0.6, 0.6);
   glBegin(GL_QUADS);
     glVertex2i(       0,         0);
     glVertex2i(winWidth,         0);
     glVertex2i(winWidth, winHeight);
     glVertex2i(       0, winHeight);
   glEnd();
   glutSwapBuffers();
   // Fractal window.
   MAIN_WINDOW.DRAW();
   MAIN_WINDOW.CLEAR();
   // Directory window.
   TEXT_8_X_13_WHITE( 10, 555, "Directory");
   FILE_WINDOW.DRAW();
   FILE_WINDOW.CLEAR();
   // Parameters window.
   TEXT_8_X_13_WHITE( 10, 170, "Parameters");
   DATA_WINDOW.DRAW();
   DATA_WINDOW.CLEAR();
   DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFfracType);
   DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFdetail);
   DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFxStart);
   DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFxEnd);
   DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFxPivot);
   DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFyStart);
   DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFyEnd);
   DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFyPivot);
   // Draw control buttons.
   TEXT_8_X_13_WHITE( 10, winHeight-20, "Commands");
   LOAD_BUTTON.DRAW();
   SAVE_BUTTON.DRAW();
   EXPT_BUTTON.DRAW();
   MAKE_BUTTON.DRAW();
   COLR_BUTTON.DRAW();
   TYPE_BUTTON.DRAW();
   DETP_BUTTON.DRAW();
   DETM_BUTTON.DRAW();
   ZOMI_BUTTON.DRAW();
   ZOMO_BUTTON.DRAW();
   PIVT_BUTTON.DRAW();
   EXIT_BUTTON.DRAW_HILITE(1.0, 0.1, 0.1);
   REST_BUTTON.DRAW_HILITE(0.5, 0.5, 1.0);
   UP___BUTTON.HIDE();
   DOWN_BUTTON.HIDE();
   CNCL_BUTTON.HIDE();
   ACPT_BUTTON.HIDE();
}

/****************************************/
/* Button display control.              */
/****************************************/
void HIDE_LOAD_BUTTONS(void) {
   UP___BUTTON.HIDE();
   DOWN_BUTTON.HIDE();
   CNCL_BUTTON.HIDE();
}

void SHOW_LOAD_BUTTONS(void) {
   UP___BUTTON.SHOW();
   DOWN_BUTTON.SHOW();
   CNCL_BUTTON.SHOW();
}

void HIDE_SAVE_BUTTONS(void) {
   ACPT_BUTTON.HIDE();
   CNCL_BUTTON.HIDE();
}

void SHOW_SAVE_BUTTONS(void) {
   ACPT_BUTTON.SHOW();
   CNCL_BUTTON.SHOW();
}

void HIDE_COMMAND_BUTTONS(void) {
   LOAD_BUTTON.HIDE();
   SAVE_BUTTON.HIDE();
   EXPT_BUTTON.HIDE();
   MAKE_BUTTON.HIDE();
   COLR_BUTTON.HIDE();
   TYPE_BUTTON.HIDE();
   DETP_BUTTON.HIDE();
   DETM_BUTTON.HIDE();
   ZOMI_BUTTON.HIDE();
   ZOMO_BUTTON.HIDE();
   PIVT_BUTTON.HIDE();
   REST_BUTTON.HIDE();
}

void SHOW_COMMAND_BUTTONS(void) {
   LOAD_BUTTON.SHOW();
   SAVE_BUTTON.SHOW();
   EXPT_BUTTON.SHOW();
   MAKE_BUTTON.SHOW();
   COLR_BUTTON.SHOW();
   TYPE_BUTTON.SHOW();
   DETP_BUTTON.SHOW();
   DETM_BUTTON.SHOW();
   ZOMI_BUTTON.SHOW();
   ZOMO_BUTTON.SHOW();
   PIVT_BUTTON.SHOW();
   REST_BUTTON.SHOW();
}

/****************************************/
/* Display whole fractal.               */
/****************************************/
void DISPLAY_FRACTAL(WINDOW window) {
int    x, y;
uchar  *ptr;
   glDrawBuffer(GL_BACK);
   glPointSize(2);
   for (y = 0; y < fracHeight; y++) {
       ptr = fractal[y];
       for (x = 0; x < fracWidth; x++) {
           glColor3ub(paletteR[*ptr], paletteG[*ptr], paletteB[*ptr]);
           glBegin(GL_POINTS);
             glVertex2i(x + window.xs + 2, window.ye - y - 3);
           glEnd();
           ptr++;
       }
   }
   glPointSize(1);
   glutSwapBuffers();
}

/****************************************/
/* Display row.                         */
/****************************************/
void DISPLAY_FRACTAL_ROW(int row, WINDOW window) {
int    x;
GLint  w = (window.xs + 2);
GLint  h = (window.ye - 3 - row);
uchar  *ptr;
   ptr = fractal[row];
   for (x = 0; x < fracWidth; x++) {
       glColor3ub(paletteR[*ptr], paletteG[*ptr], paletteB[*ptr]);
       glBegin(GL_POINTS);
         glVertex2i(w, h);
       glEnd();
       ptr++;
       w++;
   }
}

/****************************************/
/* Display fractal data.                */
/****************************************/
void DISPLAY_FRACTAL_DATA(WINDOW window, int tag) {
GLint  l = window.xs + 5;
GLint  x = window.xs + 73;
GLint  y = window.ye - 20;
char   text[80];
   glDrawBuffer(GL_BACK);
   glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
   glColor3f(0.0, 0.0, 0.0);
   if (tag == DEFfracType) {
      glBegin(GL_QUADS);
        glVertex2i(window.xs, y-4);
        glVertex2i(window.xe, y-4);
        glVertex2i(window.xe, y+12);
        glVertex2i(window.xs, y+12);
      glEnd();
      TEXT_8_X_13_WHITE(l, y, "Type---- ");
      TEXT_8_X_13_WHITE(x, y, fracType);
   } else if (tag == DEFdetail) {
      y -= 20;
      glBegin(GL_QUADS);
        glVertex2i(window.xs, y-4);
        glVertex2i(window.xe, y-4);
        glVertex2i(window.xe, y+12);
        glVertex2i(window.xs, y+12);
      glEnd();
      TEXT_8_X_13_WHITE(l, y, "Detail--    ");
      sprintf(text, " %d", detail);
      TEXT_8_X_13_WHITE(x, y, text);
   } else if (tag == DEFxStart) {
      y -= 40;
      glBegin(GL_QUADS);
        glVertex2i(window.xs, y-4);
        glVertex2i(window.xe, y-4);
        glVertex2i(window.xe, y+12);
        glVertex2i(window.xs, y+12);
      glEnd();
      TEXT_8_X_13_WHITE(l, y, "x Start-");
      sprintf(text, "%11.8lf", xStart);
      TEXT_8_X_13_WHITE(x, y, text);
   } else if (tag == DEFxEnd) {
      y -= 55;
      glBegin(GL_QUADS);
        glVertex2i(window.xs, y-4);
        glVertex2i(window.xe, y-4);
        glVertex2i(window.xe, y+12);
        glVertex2i(window.xs, y+12);
      glEnd();
      TEXT_8_X_13_WHITE(l, y, "x End---");
      sprintf(text, "%11.8lf", xEnd);
      TEXT_8_X_13_WHITE(x, y, text);
   } else if (tag == DEFxPivot) {
      y -= 70;
      glBegin(GL_QUADS);
        glVertex2i(window.xs, y-4);
        glVertex2i(window.xe, y-4);
        glVertex2i(window.xe, y+12);
        glVertex2i(window.xs, y+12);
      glEnd();
      TEXT_8_X_13_WHITE(l, y, "x Pivot-");
      sprintf(text, "%11.8lf", xPivot);
      TEXT_8_X_13_WHITE(x, y, text);
   } else if (tag == DEFyStart) {
      y -= 90;
      glBegin(GL_QUADS);
        glVertex2i(window.xs, y-4);
        glVertex2i(window.xe, y-4);
        glVertex2i(window.xe, y+12);
        glVertex2i(window.xs, y+12);
      glEnd();
      TEXT_8_X_13_WHITE(l, y, "y Start-");
      sprintf(text, "%11.8lf", yStart);
      TEXT_8_X_13_WHITE(x, y, text);
   } else if (tag == DEFyEnd) {
      y -= 105;
      glBegin(GL_QUADS);
        glVertex2i(window.xs, y-4);
        glVertex2i(window.xe, y-4);
        glVertex2i(window.xe, y+12);
        glVertex2i(window.xs, y+12);
      glEnd();
      TEXT_8_X_13_WHITE(l, y, "y End---");
      sprintf(text, "%11.8lf", yEnd);
      TEXT_8_X_13_WHITE(x, y, text);
   } else if (tag == DEFyPivot) {
      y -= 125;
      glBegin(GL_QUADS);
        glVertex2i(window.xs, y-4);
        glVertex2i(window.xe, y-4);
        glVertex2i(window.xe, y+12);
        glVertex2i(window.xs, y+12);
      glEnd();
      TEXT_8_X_13_WHITE(l, y, "y Pivot-");
      sprintf(text, "%11.8lf", yPivot);
      TEXT_8_X_13_WHITE(x, y, text);
   }
   glutSwapBuffers();
}

/****************************************/
/* Generate pallete.                    */
/****************************************/
void DISPLAY_PALLETE(void) {
int    i;
uchar  *ptrR, *ptrG, *ptrB;
uchar  LIN, ILIN, SQRT, LOG10;
uchar  SIN, COS, SIN2, COS2;
   paletteR[255] = 255;
   paletteG[255] = 255;
   paletteB[255] = 255;
   for (i = 0; i < 255; i++) {
       LIN   = (uchar)i;
       ILIN  = (uchar)(255 - i);
       SQRT  = (uchar)floor(sqrt((double)i) * 255.0 / 15.9374);
       LOG10 = (uchar)floor(log10(i) * 255.0 / 2.4049);
       SIN   = (uchar)floor(sin((double)i/81.0) * 255.0);
       COS   = (uchar)floor(1.0 + cos((double)i/81.0) * 127.0);
       SIN2  = (uchar)floor(1.0 + sin((double)i/40.0) * 127.0);
       COS2  = (uchar)floor(1.0 + cos((double)i/40.0) * 127.0);
       // Gray scales.
       if (palette == 0) {
          paletteR[i] = LIN;
          paletteG[i] = LIN;
          paletteB[i] = LIN;
       } else if (palette == 1) {
          paletteR[i] = SQRT;
          paletteG[i] = SQRT;
          paletteB[i] = SQRT;
       } else if (palette == 2) {
          paletteR[i] = LOG10;
          paletteG[i] = LOG10;
          paletteB[i] = LOG10;

       // Linear color scales.
       } else if (palette == 3) {
          paletteR[i] = ILIN;
          paletteG[i] = LIN;
          paletteB[i] = LIN;
       } else if (palette == 4) {
          paletteR[i] = LIN;
          paletteG[i] = ILIN;
          paletteB[i] = LIN;
       } else if (palette == 5) {
          paletteR[i] = LIN;
          paletteG[i] = LIN;
          paletteB[i] = ILIN;

       // Non-linear color scales.
       } else if (palette == 6) {
          paletteR[i] = ILIN;
          paletteG[i] = LOG10;
          paletteB[i] = LIN;
       } else if (palette == 7) {
          paletteR[i] = LIN;
          paletteG[i] = ILIN;
          paletteB[i] = LOG10;
       } else if (palette == 8) {
          paletteR[i] = LOG10;
          paletteG[i] = LIN;
          paletteB[i] = ILIN;
       } else if (palette == 9) {
          paletteR[i] = ILIN;
          paletteG[i] = SQRT;
          paletteB[i] = LIN;
       } else if (palette == 10) {
          paletteR[i] = LIN;
          paletteG[i] = ILIN;
          paletteB[i] = SQRT;
       } else if (palette == 11) {
          paletteR[i] = SQRT;
          paletteG[i] = LIN;
          paletteB[i] = ILIN;
       } else if (palette == 12) {
          paletteR[i] = ILIN;
          paletteG[i] = SQRT;
          paletteB[i] = LOG10;
       } else if (palette == 13) {
          paletteR[i] = LOG10;
          paletteG[i] = ILIN;
          paletteB[i] = SQRT;
       } else if (palette == 14) {
          paletteR[i] = SQRT;
          paletteG[i] = LOG10;
          paletteB[i] = ILIN;
       } else if (palette == 15) {
          paletteR[i] = LIN;
          paletteG[i] = SIN;
          paletteB[i] = LOG10;
       } else if (palette == 16) {
          paletteR[i] = LOG10;
          paletteG[i] = LIN;
          paletteB[i] = SIN;
       } else if (palette == 17) {
          paletteR[i] = SIN;
          paletteG[i] = LOG10;
          paletteB[i] = LIN;
       } else if (palette == 18) {
          paletteR[i] = LIN;
          paletteG[i] = SIN;
          paletteB[i] = SQRT;
       } else if (palette == 19) {
          paletteR[i] = SQRT;
          paletteG[i] = LIN;
          paletteB[i] = SIN;
       } else if (palette == 20) {
          paletteR[i] = SIN;
          paletteG[i] = SQRT;
          paletteB[i] = LIN;
       } else if (palette == 21) {
          paletteR[i] = LIN;
          paletteG[i] = SIN;
          paletteB[i] = COS;
       } else if (palette == 22) {
          paletteR[i] = COS;
          paletteG[i] = LIN;
          paletteB[i] = SIN;
       } else if (palette == 23) {
          paletteR[i] = SIN;
          paletteG[i] = COS;
          paletteB[i] = LIN;
       } else if (palette == 24) {
          paletteR[i] = SQRT;
          paletteG[i] = SIN;
          paletteB[i] = COS;
       } else if (palette == 25) {
          paletteR[i] = COS;
          paletteG[i] = SQRT;
          paletteB[i] = SIN;
       } else if (palette == 26) {
          paletteR[i] = SIN;
          paletteG[i] = COS;
          paletteB[i] = SQRT;
       } else if (palette == 27) {
          paletteR[i] = LOG10;
          paletteG[i] = SIN;
          paletteB[i] = COS;
       } else if (palette == 28) {
          paletteR[i] = COS;
          paletteG[i] = LOG10;
          paletteB[i] = SIN;
       } else if (palette == 29) {
          paletteR[i] = SIN;
          paletteG[i] = COS;
          paletteB[i] = LOG10;
       } else if (palette == 30) {
          paletteR[i] = LOG10;
          paletteG[i] = SIN2;
          paletteB[i] = COS2;
       } else if (palette == 31) {
          paletteR[i] = COS2;
          paletteG[i] = LOG10;
          paletteB[i] = SIN2;
       } else if (palette == 32) {
          paletteR[i] = SIN2;
          paletteG[i] = COS2;
          paletteB[i] = LOG10;
       } else if (palette == 33) {
          paletteR[i] = COS2;
          paletteG[i] = SIN2;
          paletteB[i] = COS2;
       } else if (palette == 34) {
          paletteR[i] = COS2;
          paletteG[i] = COS2;
          paletteB[i] = SIN2;
       } else if (palette == 35) {
          paletteR[i] = SIN2;
          paletteG[i] = COS2;
          paletteB[i] = COS2;
       }
   }
}


/***********************************************************/
/* FILE I/O ROUTINES                                       */
/***********************************************************/
/****************************************/
/* File scan.                     	*/
/****************************************/
int  SCAN_FOR_IMAGES(const char Extension[4]) {
int    i;
char   sysCmd[80];
FILE   *disk;
   strcpy(sysCmd, "ls *");
   strcat(sysCmd, Extension);
   strcat(sysCmd, " > fileList");
   system(sysCmd);
   if ((disk = fopen("fileList", "r")) != NULL) {
      for (i = 0; i < maxNames; i++) {
          fscanf(disk, "%s", fileNames[i]);
          if (feof(disk)) break;
      }
   }
   fclose(disk);
   system("rm fileList");
   if (i == 0) return 0;
          else return --i;
}

/****************************************/
/* File scan display.                   */
/****************************************/
void SCAN_FILE_DISPLAY(int start, int fractals) {
int    i, j = start;
   FILE_WINDOW.CLEAR();
   glColor3f(1.0, 1.0, 1.0);
   for (i = 15; i <= 300; i+=20) {
       TEXT_8_X_13_WHITE(FILE_WINDOW.xs+5, FILE_WINDOW.ye-i, fileNames[j]);
       if (j == fractals) break;
       j++;
   }
   glutSwapBuffers();
}

/****************************************/
/* Load fractal.                     	*/
/****************************************/
int  LOAD_FRACTAL(char fileName[80]) {
int    x, y, k = 0;
int    pmax;
char   ch;
char   type[3];
FILE   *disk;
uchar  *ptr;
   disk = fopen(fileName, "rb");
   // Fractal data.
   fscanf(disk, "%s",  fracType);
   if (strcmp(fracType, "FRACRAM")) return 0;
   fscanf(disk, "%s",  fracType);
   if (!strcmp(fracType, "Mandelbrot")) fType = 0;
   else if (!strcmp(fracType, "Julia")) fType = 1;
   fscanf(disk, "%d",  &detail);
   fscanf(disk, "%lf", &xStart);
   fscanf(disk, "%lf", &xEnd);
   fscanf(disk, "%lf", &xPivot);
   fscanf(disk, "%lf", &yStart);
   fscanf(disk, "%lf", &yEnd);
   fscanf(disk, "%lf", &yPivot);
   // Fractal image.
   fscanf(disk, "%s", type);
   fread(&ch, 1, 1, disk); 
   fscanf(disk, "%d", &fracWidth);
   fscanf(disk, "%d", &fracHeight);
   fscanf(disk, "%d", &pmax);
   fread(&ch, 1, 1, disk);
   for (y = 0; y < fracHeight; y++) {
       ptr = fractal[y];
       for (x = 0; x < fracWidth; x++) {
           *ptr = getc(disk);
           ptr++;
       }
   }
   fclose(disk);
   return 1;
}

/****************************************/
/* Save fractal.                     	*/
/****************************************/
int  SAVE_FRACTAL(char fileName[80]) {
int    x, y;
uchar  *ptr;
FILE   *disk;
   disk = fopen(fileName, "wb");
   // Fractal data.
   fprintf(disk, "FRACRAM\n");
   fprintf(disk, "%s\n",  fracType);
   fprintf(disk, "%d\n",  detail);
   fprintf(disk, "%lf\n", xStart);
   fprintf(disk, "%lf\n", xEnd);
   fprintf(disk, "%lf\n", xPivot);
   fprintf(disk, "%lf\n", yStart);
   fprintf(disk, "%lf\n", yEnd);
   fprintf(disk, "%lf\n", yPivot);
   // Fractal image.
   fprintf(disk, "P5\n");
   fprintf(disk, "%d %d\n", fracWidth, fracHeight);
   fprintf(disk, "255\n");
   for (y = 0; y < fracHeight; y++) {
       ptr = fractal[y];
       for (x = 0; x < fracWidth; x++) {
           fprintf(disk, "%c", *ptr);
           ptr++;
       }
   } 
   fclose(disk);
}

/****************************************/
/* Export fractal.                     	*/
/****************************************/
void EXPORT_FRACTAL(char fileName[80]) {
int    x, y;
uchar  *ptr, *ptd;
uchar  data[800*800*3];
   ptd = data;
   for (y = 0; y < fracHeight; y++) {
       ptr = fractal[y];
       for (x = 0; x < fracWidth; x++) {
           *ptd = paletteR[*ptr];
           ptd++;
           *ptd = paletteG[*ptr];
           ptd++;
           *ptd = paletteB[*ptr];
           ptd++;
           ptr++;
       }
   }
   PGM_FILE_WRITE_1D(fileName, data, fracWidth, fracHeight, 1);
}

/***********************************************************/
/* FRACTAL GENERATION ROUTINE ALGORITHM CORES              */
/***********************************************************/
uchar *MANDELBROT_CORE(int y, uchar *ptr) {
int    x, color;
double xx, yy, xs, ys, xy, cx, cy;
   cy  = yStart + (y * yScale);
   for (x = 0; x < fracWidth; x++) {
       cx    = xStart + (x * xScale);
       xx    = 0.0;
       yy    = 0.0;
       xs    = 0.0;
       ys    = 0.0;
       xy    = 0.0;
       color = -1;
       do {
          xs = (xx * xx);
          ys = (yy * yy);
          xy = (xx * yy);
          xx = (xs - ys + cx);
          yy = (xy + xy + cy);
          color++;
       } while ((color < detail) && ((xs + ys) <= 4.0));
       *ptr = ((color * 255) / detail);
       ptr++;
   }
   return ptr;
}

/***********************************************************/
/* UNIPROCESS FRACTAL GENERATION ROUTINES                  */
/***********************************************************/
/****************************************/
/* Mandelbrot.                          */
/****************************************/
void UNIPROCESS_MANDELBROT_GENERATE(WINDOW window) {
int y;
uchar *ptr;
   glDrawBuffer(GL_FRONT_AND_BACK);
   glPointSize(2);
   xScale = (xEnd - xStart) / (double)fracWidth;
   yScale = (yEnd - yStart) / (double)fracHeight;
   for (y = 0; y < fracHeight; y++) {
       ptr = MANDELBROT_CORE(y, fractal[y]);
       DISPLAY_FRACTAL_ROW(y, window);
   } 
   glPointSize(1);
}



/*
void UNIPROCESS_MANDELBROT_GENERATE(WINDOW window) {
int    x, y, color;
double xx, yy, xs, ys, xy, cx, cy;
uchar  *ptr;
   glDrawBuffer(GL_FRONT_AND_BACK);
   glPointSize(2);
   xScale = (xEnd - xStart) / (double)fracWidth;
   yScale = (yEnd - yStart) / (double)fracHeight;
   for (y = 0; y < fracHeight; y++) {
       ptr = fractal[y];
       cy  = yStart + (y * yScale);
       for (x = 0; x < fracWidth; x++) {
           cx    = xStart + (x * xScale);
           xx    = 0.0;
           yy    = 0.0;
           xs    = 0.0;
           ys    = 0.0;
           xy    = 0.0;
           color = -1;
           do {
              xs = (xx * xx);
              ys = (yy * yy);
              xy = (xx * yy);
              xx = (xs - ys + cx);
              yy = (xy + xy + cy);
              color++;
           } while ((color < detail) && ((xs + ys) <= 4.0));
           *ptr = ((color * 255) / detail);
           ptr++;
           DISPLAY_FRACTAL_ROW(y, window);
       } 
   }
   glPointSize(1);
}
*/

/****************************************/
/* Julia.                               */
/****************************************/
void UNIPROCESS_JULIA_GENERATE(WINDOW window) {
int    x, y, color;
double xx, yy, xs, ys, xy, cx, cy;
uchar  *ptr;
   glDrawBuffer(GL_FRONT_AND_BACK);
   glPointSize(2);
   xScale = (xEnd - xStart) / (double)fracWidth;
   yScale = (yEnd - yStart) / (double)fracHeight;
   for (y = 0; y < fracHeight; y++) {
       ptr = fractal[y];
       for (x = 0; x < fracWidth; x++) {
           yy    = yStart + (y * yScale);
           xx    = xStart + (x * xScale);
           color = -1;
           do {
               xs = (xx * xx);
               ys = (yy * yy);
               xy = (xx * yy);
               xx = (xs - ys + xPivot);
               yy = (xy + xy + yPivot);
               color++;
           } while ((color < detail) && ((xs + ys) <= 4.0));
           *ptr = ((color * 255) / detail);
           ptr++;
           ptr++;
           DISPLAY_FRACTAL_ROW(y, window);
       } 
   }
   glPointSize(1);
}

/****************************************/
/* Mandelbrot Dragon.                   */
/****************************************/
void UNIPROCESS_DRAGON_GENERATE(WINDOW window) {
int    x, y, j, color;
double xx, yy, xs, ys, xt, yt, cx, cy, ct;
uchar  *ptr;
   glDrawBuffer(GL_FRONT_AND_BACK);
   glPointSize(2);
   xScale = (xEnd - xStart) / (double)fracWidth;
   yScale = (yEnd - yStart) / (double)fracHeight;
   for (y = 0; y < fracHeight; y++) {
       ptr = fractal[y];
       cy  = yStart + (y * yScale);
       for (x = 0; x < fracWidth; x++) {
           cx    = xStart + (x * xScale);
           xx    = 0.5;
           yy    = 0.0;
           xs    = 0.0;
           ys    = 0.0;
           ct    = cy;
           color = -1;
           do {
              xs = (xx * xx);
              ys = (yy * yy);
              xt = (ys - xs + xx);
              yt = ((xx * yy * 2) - yy);
              xx = ((cx * xt) + (ct * yt));
              yy = ((ct * xt) - (cx * yt));
              color++;
           } while ((color < detail) && ((xs + ys) <= 4.0));
           *ptr = ((color * 255) / detail);
           ptr++;
           DISPLAY_FRACTAL_ROW(y, window);
       } 
   }
   glPointSize(1);
}

/****************************************/
/* Julia Dragon.                        */
/****************************************/
void UNIPROCESS_J_DRAGON_GENERATE(WINDOW window) {
int    x, y, j, color;
double xx, yy, xs, ys, t1, t2, cx, cy;
uchar  *ptr;
   glDrawBuffer(GL_FRONT_AND_BACK);
   glPointSize(2);
   xScale = (xEnd - xStart) / (double)fracWidth;
   yScale = (yEnd - yStart) / (double)fracHeight;
   for (y = 0; y < fracHeight; y++) {
       ptr = fractal[y];
       cy  = yStart + (y * yScale);
       for (x = 0; x < fracWidth; x++) {
           cx    = xStart + (x * xScale);
           xx    = cx;
           yy    = cy;
           xs    = 0.0;
           ys    = 0.0;
           color = -1;
           do {
              xs = (xx * xx);
              ys = (yy * yy);
              t1 = ((ys - xs) + xx);
              t2 = ((xx * yy * 2) - yy);
              xx = ((xPivot * t1) + (yPivot * t2));
              yy = ((yPivot * t1) - (xPivot * t2));
              color++;
           } while ((color < detail) && ((xs + ys) <= 4.0));
           *ptr = ((color * 255) / detail);
           ptr++;
           DISPLAY_FRACTAL_ROW(y, window);
       } 
   }
   glPointSize(1);
}

/***********************************************************/
/* THREADED FRACTAL GENERATION ROUTINES                    */
/***********************************************************/
/****************************************/
/* Generate MANDELBROT.                 */
/****************************************/
void *MANDELBROT(void *arg) {
int y, exitCode = 1;
uchar *ptr;
struct thread_data *params;
   params = ((thread_data*)arg);
   xScale = (xEnd - xStart) / (double)fracWidth;
   yScale = (yEnd - yStart) / (double)fracHeight;
   for (y = 0; y < fracHeight; y++) {
       ptr = MANDELBROT_CORE(params->row, fractal[params->row]);
   } 
   params->status = 1;
   pthread_exit(&exitCode);
}


/*
void *MANDELBROT(void *arg) {
int    x, y, color;
int    exitCode = 1;
double xx, yy, xs, ys, xy, cx, cy;
uchar  *ptr;
struct thread_data *params;
    params = ((thread_data*)arg);
    xScale = (xEnd - xStart) / (double)fracWidth;
    yScale = (yEnd - yStart) / (double)fracHeight;
    y   = params->row;
    cy  = yStart + (y * yScale);
    ptr = fractal[y];
    for (x = 0; x < fracWidth; x++) {
        cx    = xStart + (x * xScale);
        xx    = 0.0;
        yy    = 0.0;
        color = -1;
        do {
            xs = (xx * xx);
            ys = (yy * yy);
            xy = (xx * yy);
            xx = (xs - ys + cx);
            yy = (xy + xy + cy);
            color++;
        } while ((color < detail) && ((xs + ys) <= 4.0));
        *ptr = ((color * 255) / detail);
        ptr++;
    }
    params->status = 1;
    pthread_exit(&exitCode);
}
*/

/****************************************/
/* Generate JULIA.                      */
/****************************************/
void *JULIA(void *arg) {
int    x, y, color;
int    exitCode = 1;
double xx, yy, xs, ys, xy,cx, cy;
uchar  *ptr;
struct thread_data *params;
    params = ((thread_data*)arg);
    xScale = (xEnd - xStart) / (double)fracWidth;
    yScale = (yEnd - yStart) / (double)fracHeight;
    y   = params->row;
    ptr = fractal[y];
    for (x = 0; x < fracWidth; x++) {
        cx    = xPivot;
        cy    = yPivot;
        yy    = yStart + (y * yScale);
        xx    = xStart + (x * xScale);
        color = -1;
        do {
            xs = (xx * xx);
            ys = (yy * yy);
            xy = (xx * yy);
            xx = (xs - ys + cx);
            yy = (xy + xy + cy);
            color++;
        } while ((color < detail) && ((xs + ys) <= 4.0));
        *ptr = ((color * 255) / detail);
        ptr++;
    }
    params->status = 1;
    pthread_exit(&exitCode);
}

/****************************************/
/* Generate fractal using N threads on  */
/* a single machine.                    */
/****************************************/
void THREAD_FRACTAL_GENERATE(WINDOW window) {
int    x, y, t;
int    row, retn, send;
pthread_t fractalRow[threads];
struct thread_data bounds[threads];
   glDrawBuffer(GL_FRONT_AND_BACK);
   glPointSize(2);
    // Divide into rows and send each to a thread.
   row = 0;
   for (t = 0; t < threads; t++) {
       bounds[t].row    = row;
       bounds[t].status = 0;
       if (fType == 0) pthread_create(&fractalRow[t], NULL, MANDELBROT, (void *)&bounds[t]); 
       if (fType == 1) pthread_create(&fractalRow[t], NULL, JULIA, (void *)&bounds[t]); 
       row++;
   }
   // Wait for any thread to complete and resend new data.
   t    = 0;
   retn = 0;
   send = 1;
   while (1) {
      if (bounds[t].status == 1) {
         bounds[t].status = 0;
         pthread_detach(fractalRow[t]);
         retn++;
         DISPLAY_FRACTAL_ROW(bounds[t].row, window);
         // Keep sending data?
         if (send == 1) {
            bounds[t].row = row;
            if (fType == 0) pthread_create(&fractalRow[t], NULL, MANDELBROT, (void *)&bounds[t]); 
            if (fType == 1) pthread_create(&fractalRow[t], NULL, JULIA, (void *)&bounds[t]); 
            row++;
         }
         if (row  == fracHeight) send = 0;
         if (retn == fracHeight) break;
      }
      t++;
      if (t == threads) t = 0;
   }
   glPointSize(1);
}


/***********************************************************/
/* CALLBACK FUNTIONS.                                      */
/***********************************************************/
/****************************************/
/* Mouse control.                       */
/****************************************/
void myMouse(int button, int state, int x, int y) {
int  filePicked;
char fileName[80];
time_t start, stop;
   y = (winHeight - y);
   if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN) {
      glFinish();
      /**********************************/
      /* Load fractal from disk.  	*/
      /**********************************/
      if (LOAD_BUTTON.SELECTED(x, y)) { 
         LOAD_BUTTON.CLICK();
         lockMenu  = 1;
         SHOW_LOAD_BUTTONS();
         scanFlag  = 1;
         scanStart = 0;
         fractals  = SCAN_FOR_IMAGES("frc");
         SCAN_FILE_DISPLAY(scanStart, fractals);
         return;
      }
      /**********************************/
      /* Save fractal to disk.          */
      /**********************************/
      if (SAVE_BUTTON.SELECTED(x, y)) { 
         if (lockMenu == 1 || noFractal == 1) return;
         SAVE_BUTTON.CLICK();
         lockMenu  = 1;
         strcpy(fileNames[0], "");
         SHOW_SAVE_BUTTONS();
         FILE_WINDOW.CLEAR();
         TEXT_8_X_13_WHITE(FILE_WINDOW.xs+5, FILE_WINDOW.ye-15 , "Enter filename:");
         scanFlag  = -1;
         return;
      }
      /**********************************/
      /* Export fractal to PPM file.    */
      /**********************************/
      if (EXPT_BUTTON.SELECTED(x, y)) { 
         if (lockMenu == 1 || noFractal == 1) return;
         EXPT_BUTTON.CLICK();
         lockMenu  = 1;
         strcpy(fileNames[0], "");
         SHOW_SAVE_BUTTONS();
         FILE_WINDOW.CLEAR();
         TEXT_8_X_13_WHITE(FILE_WINDOW.xs+5, FILE_WINDOW.ye-15 , "Enter filename:");
         scanFlag  = -2;
         return;
      }

      /**********************************/
      /* Select fractal type.           */
      /**********************************/
      if (TYPE_BUTTON.SELECTED(x, y)) { 
         if (lockMenu == 1) return;
         TYPE_BUTTON.CLICK();
         fType++;
         if (fType > MaxfracType) fType = 0;
         if (fType == 0) strcpy(fracType, "Mandelbrot");   
         else if (fType == 1) strcpy(fracType, "Julia");   
         else if (fType == 2) strcpy(fracType, "Dragon"); 
         else if (fType == 3) strcpy(fracType, "J-Dragon");     
         DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFfracType);
      }

      /**********************************/
      /* Generate fractal.              */
      /**********************************/
      if (MAKE_BUTTON.SELECTED(x, y)) { 
         if (lockMenu == 1) return;
         MAKE_BUTTON.CLICK();
         MAKE_BUTTON.DRAW();
         MAIN_WINDOW.CLEAR();
         start = time(NULL);
         if (processType == 1) {
            if (fType == 0) UNIPROCESS_MANDELBROT_GENERATE(MAIN_WINDOW);
            if (fType == 1) UNIPROCESS_JULIA_GENERATE(MAIN_WINDOW);
            if (fType == 2) UNIPROCESS_DRAGON_GENERATE(MAIN_WINDOW);
            if (fType == 3) UNIPROCESS_J_DRAGON_GENERATE(MAIN_WINDOW);
         }
         if (processType == 2) THREAD_FRACTAL_GENERATE(MAIN_WINDOW);
         stop = time(NULL);
         printf("Run time (sec) = %d.\n", (stop - start));
         noFractal = 0;
         return;
      }

      /**********************************/
      /* Select palette.                */
      /**********************************/
      if (COLR_BUTTON.SELECTED(x, y)) { 
         if (lockMenu == 1 || noFractal == 1) return;
         COLR_BUTTON.CLICK();
         palette++;
         if (palette > 35) palette = 0;
         DISPLAY_PALLETE();
         DISPLAY_FRACTAL(MAIN_WINDOW);
         return;
      }

      /**********************************/
      /* Increase fractal detail.       */        
      /**********************************/
      if (DETP_BUTTON.SELECTED(x, y)) { 
         if (lockMenu == 1) return;
         DETP_BUTTON.CLICK();
         detail+=50;
         if (detail > 5000) detail = 5000;
         DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFdetail);
         MAKE_BUTTON.DRAW_HILITE(1.0, 1.0, 1.0);
         return;
      }

      /**********************************/
      /* Decrease fractal detail.       */        
      /**********************************/
      if (DETM_BUTTON.SELECTED(x, y)) { 
         if (lockMenu == 1) return;
         DETM_BUTTON.CLICK();
         detail-=50;
         if (detail < 50) detail = 50;
         DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFdetail);
         MAKE_BUTTON.DRAW_HILITE(1.0, 1.0, 1.0);
         return;
      }

      /**********************************/
      /* Zoom in.                       */        
      /**********************************/
      if (ZOMI_BUTTON.SELECTED(x, y)) { 
         if (lockMenu == 1 || noFractal == 1) return;
         ZOMI_BUTTON.CLICK();
         lockMenu  = 1;
         HIDE_COMMAND_BUTTONS();
         glutSetCursor(GLUT_CURSOR_CROSSHAIR);
         zoomFlag = 1;
         return;
      }

      /**********************************/
      /* Zoom out.                      */        
      /**********************************/
      if (ZOMO_BUTTON.SELECTED(x, y)) { 
         if (lockMenu == 1 || noFractal == 1) return;
         ZOMO_BUTTON.CLICK();
         xStart    = -5.0;
         xEnd      =  5.0;
         xScale    = 10.0 / (double)fracWidth;
         yStart    = -5.0;
         yEnd      =  5.0;
         yScale    = 10.0 / (double)fracHeight;
         DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFxStart);
         DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFxEnd);
         DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFyStart);
         DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFyEnd);
         MAKE_BUTTON.DRAW_HILITE(1.0, 1.0, 1.0);
         return;
      }

      /**********************************/
      /* Select pivot.                  */        
      /**********************************/
      if (PIVT_BUTTON.SELECTED(x, y)) { 
         if (lockMenu == 1 || noFractal == 1) return;
         PIVT_BUTTON.CLICK();
         lockMenu  = 1;
         HIDE_COMMAND_BUTTONS();
         glutSetCursor(GLUT_CURSOR_CROSSHAIR);
         zoomFlag = -1;
         return;
      }

      /**********************************/
      /* Reset parameters.              */        
      /**********************************/
      if (REST_BUTTON.SELECTED(x, y)) { 
         if (lockMenu == 1) return;
         REST_BUTTON.CLICK();
         palette   = 0;
         detail    = 50;
         xStart    = -5.0;
         xEnd      =  5.0;
         xPivot    =  0.0;
         xScale    = 10.0 / (double)fracWidth;
         yStart    = -5.0;
         yEnd      =  5.0;
         yPivot    =  0.0;
         yScale    = 10.0 / (double)fracHeight;
         strcpy(fracType, "Mandelbrot");   
         DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFfracType);
         DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFdetail);
         DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFxStart);
         DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFxEnd);
         DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFxPivot);
         DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFyStart);
         DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFyEnd);
         DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFyPivot);
         MAIN_WINDOW.CLEAR();
      }

      /**********************************/
      /* Exit.                          */        
      /**********************************/
      if (EXIT_BUTTON.SELECTED(x, y)) { 
         EXIT_BUTTON.CLICK();
         glutDestroyWindow(window);
         exit(0);                   
      }

      /**********************************/
      /* Directory listing control.     */        
      /**********************************/
      if (UP___BUTTON.SELECTED(x, y)) { 
         UP___BUTTON.CLICK();
         if (scanStart <= (fractals-15)) {
            scanStart++;
            SCAN_FILE_DISPLAY(scanStart, fractals);
         }
         return;
      }
      if (DOWN_BUTTON.SELECTED(x, y)) { 
         DOWN_BUTTON.CLICK();
         if (scanStart > 0) {
            scanStart--;
            SCAN_FILE_DISPLAY(scanStart, fractals);
         }
         return;
      }
      if (CNCL_BUTTON.SELECTED(x, y)) { 
         CNCL_BUTTON.CLICK();
         FILE_WINDOW.CLEAR();
         if (scanFlag == 1) {
            HIDE_LOAD_BUTTONS();
            HIDE_SAVE_BUTTONS();
         }
         if (scanFlag == -1) {
            HIDE_SAVE_BUTTONS();
            HIDE_SAVE_BUTTONS();
         }
         lockMenu = 0;
         return;
      }
      if (ACPT_BUTTON.SELECTED(x, y)) { 
         ACPT_BUTTON.CLICK();
         FILE_WINDOW.CLEAR();
         HIDE_SAVE_BUTTONS();
         if (scanFlag == -1) {
            strcat(fileNames[0], ".frc");
            SAVE_FRACTAL(fileNames[0]);
         }
         if (scanFlag == -2) {
            strcat(fileNames[0], ".ppm");
            EXPORT_FRACTAL(fileNames[0]);
         }
         lockMenu = 0;
         return;
      }
      // Filename selection.
      if (x >= 10 && x <= 170 && scanFlag == 1) {
         filePicked = scanStart + ((785 - y) / 20) - 15;
         if (filePicked < 0 || filePicked > fractals) return;
         strcpy(fileName, fileNames[filePicked]);
         if (LOAD_FRACTAL(fileName) == 1) {
            MAIN_WINDOW.CLEAR();
            DISPLAY_FRACTAL(MAIN_WINDOW);
            DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFfracType);
            DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFdetail);
            DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFxStart);
            DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFxEnd);
            DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFxPivot);
            DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFyStart);
            DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFyEnd);
            DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFyPivot);
         }
         scanFlag = 0;
         FILE_WINDOW.CLEAR();
         HIDE_LOAD_BUTTONS();
         noFractal= 0;
         lockMenu = 0;
         xScale   = (xEnd - xStart) / (double)fracWidth;
         yScale   = (yEnd - yStart) / (double)fracHeight;
         return;
      }
   }

   /**********************************/
   /* Zoom control.                  */        
   /**********************************/
   if (zoomFlag == 1 && button == GLUT_LEFT_BUTTON && state == GLUT_DOWN) {
      glFinish();
      xTemp    = xStart + ((x - MAIN_WINDOW.xs) * xScale);
      yTemp    = yStart + ((MAIN_WINDOW.ye - y) * yScale);
      xs       = x;
      ys       = y;
      zoomFlag = 2;
      glutSetCursor(GLUT_CURSOR_NONE);
   }
   if (zoomFlag == 2 && button == GLUT_LEFT_BUTTON && state == GLUT_UP) {
      glFinish();
      glDrawBuffer(GL_BACK);
      glPolygonMode(GL_FRONT, GL_LINE);
      glColor3ub(128, 128, 128);
      glBegin(GL_QUADS);
        glVertex2i(xs, ys);
        glVertex2i(xs, ye);
        glVertex2i(xe, ye);
        glVertex2i(xe, ys);
      glEnd();
      glutSwapBuffers();
      xEnd     = xStart + ((x - MAIN_WINDOW.xs) * xScale);
      yEnd     = yStart + ((MAIN_WINDOW.ye - y) * yScale);
      xStart   = xTemp;
      yStart   = yTemp;
      zoomFlag = 0;
      DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFxStart);
      DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFxEnd);
      DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFyStart);
      DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFyEnd);
      SHOW_COMMAND_BUTTONS();
      lockMenu = 0;
      glutSetCursor(GLUT_CURSOR_INFO);
      MAKE_BUTTON.DRAW_HILITE(1.0, 1.0, 1.0);
   }

   /**********************************/
   /* Pivot control.                 */        
   /**********************************/
   if (zoomFlag == -1 && button == GLUT_LEFT_BUTTON && state == GLUT_DOWN) {
      glFinish();
      glDrawBuffer(GL_BACK);
      glPointSize(10);
      glColor3ub(128, 128, 128);
      glBegin(GL_POINTS);
        glVertex2i(x, y);
      glEnd();
      glPointSize(1);
      glutSwapBuffers();
      xPivot   = xStart + ((x - MAIN_WINDOW.xs) * xScale);
      yPivot   = yStart + ((MAIN_WINDOW.ye - y) * yScale);
      zoomFlag = 0;
      DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFxPivot);
      DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFyPivot);
      SHOW_COMMAND_BUTTONS();
      lockMenu = 0;
      glutSetCursor(GLUT_CURSOR_INFO);
   }
}

/****************************************/
/* Mouse control.                       */
/****************************************/
void myMovingMouse(int x, int y) {
   if (zoomFlag != 2) return;
   xe = x;
   ye = (winHeight - y); 
   glDrawBuffer(GL_FRONT);
   glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
   glColor3ub(128, 128, 128);
   glBegin(GL_QUADS);
     glVertex2i(xs, ys);
     glVertex2i(xs, ye);
     glVertex2i(xe, ye);
     glVertex2i(xe, ys);
   glEnd();
   glutSwapBuffers();
}

/****************************************/
/* File prompt.                         */
/****************************************/
void myKeyboard(uchar key, int x, int y) {
char txt[2];
   if (scanFlag > -1) return;
   // Clear text area.
   glClear(GL_DEPTH_BUFFER_BIT);                 
   glDrawBuffer(GL_BACK);
   glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
   glColor3f(0.0, 0.0, 0.0);
   glBegin(GL_QUADS);
     glVertex2i(FILE_WINDOW.xs, FILE_WINDOW.ye-17);
     glVertex2i(FILE_WINDOW.xe, FILE_WINDOW.ye-17);
     glVertex2i(FILE_WINDOW.xe, FILE_WINDOW.ye-32);
     glVertex2i(FILE_WINDOW.xs, FILE_WINDOW.ye-32);
   glEnd();
   glutSwapBuffers();
   // Display filename entered.
   if ((int)key == 8) {
      fileNames[0][strlen(fileNames[0])-1] = '\0';
   } 
   if ((int)key > 31) {
      txt[0] = key;
      txt[1] = '\0';
      strcat(fileNames[0], txt);
   }
   TEXT_8_X_13_WHITE(FILE_WINDOW.xs+5, FILE_WINDOW.ye-30, fileNames[0]);
}

/****************************************/
/* Resize.                              */
/****************************************/
void myResize(int width, int height) {
   glutReshapeWindow(winWidth, winHeight);                    
}

/****************************************/
/* Display.                             */
/****************************************/
void myDisplay(void) {
   if (drawFlag == 0) {
      glutSwapBuffers();
   } else {
      DISPLAY_MAIN_PANEL();
      drawFlag = 0;
   }
}

/****************************************/
/* Init.                                */
/****************************************/
void myInit(void) {
   // initialize OpenGL parameters.
   glPointSize(1);
   glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
   glClear(GL_COLOR_BUFFER_BIT); 
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity(); 
   gluOrtho2D(0.0, (GLdouble)winWidth, 0.0, (GLdouble)winHeight);
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();     
   // Initialize prgram parameters.
   strcpy(fracType, "Mandelbrot");   
   DISPLAY_PALLETE();  
}


/***********************************************************/
/* MAIN.                                                   */
/***********************************************************/
/****************************************/
/* Main.                                */
/****************************************/
int main(int argc, char **argv) {
int  nodesPossible;
char text[8];
char title[80];
   strcpy(title, "Frac-R.A.M. (V. ");
   strcat(title, version);
   // Trap HELP option.
   if (argc == 2) { 
      system("clear");
      printf("%s\n", title);
      printf("\n");
      printf("Frac-R.A.M. is designed to support multiple modes of processing, ");
      printf("including:\n");
      printf("\n");
      printf(" 1. A single process on a single machine.\n");
      printf("    Involked by the command: fracram\n");
      printf("\n");
      printf(" 2. Multiple (N) threads on a single machine.\n");
      printf("    Involked by the command: fracram t N\n");
      printf("\n");
      exit(0);
   }
   // Mode 1 - Trap uni-process option.
   if (argc == 1) {
      processType = 1;
      threads     = 1;
      strcat(title, " - UniProcess");
   }
   // Mode 2 - Trap threaded-process option.
   if (argc == 3 && argv[1][0] == 't') {
      processType = 2;
      threads     = atoi(argv[2]);
      if (threads == 0) {
         processType = 1;
         threads     = 1;
         strcat(title, " - UniProcess");
      } else {
         strcat(title, " - Threaded x ");
         strcat(title, argv[2]);
      }
   }
   strcat(title, ") by R. Marsh.");
   
   // Start GUI.
   glutInit(&argc, argv);
   glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);  
   glutInitWindowSize(winWidth, winHeight);
   glutInitWindowPosition(10, 10);
   window = glutCreateWindow(title);
   glutSetCursor(GLUT_CURSOR_INFO);
   glutDisplayFunc(myDisplay);
   glutReshapeFunc(myResize);
   glutMouseFunc(myMouse);
   glutMotionFunc(myMovingMouse);
   glutKeyboardFunc(myKeyboard);
   myInit();
   glutMainLoop();

   return 1;
}
