#include "fracram.h"

/***********************************************************/
/* GUI ROUTINES                                            */
/***********************************************************/
/****************************************/
/* Main window display.                 */
/****************************************/
void DISPLAY_MAIN_PANEL(void) {
   glDrawBuffer(GL_BACK);
   glPolygonMode(GL_FRONT, GL_FILL);
   glColor3f(0.6, 0.6, 0.6);
   glBegin(GL_QUADS);
     glVertex2i(       0,         0);
     glVertex2i(winWidth,         0);
     glVertex2i(winWidth, winHeight);
     glVertex2i(       0, winHeight);
   glEnd();
   glutSwapBuffers();
   // Fractal window.
   MAIN_WINDOW.DRAW();
   MAIN_WINDOW.CLEAR();
   // Directory window.
   TEXT_8_X_13_WHITE( 10, 555, "Directory");
   FILE_WINDOW.DRAW();
   FILE_WINDOW.CLEAR();
   // Parameters window.
   TEXT_8_X_13_WHITE( 10, 170, "Parameters");
   DATA_WINDOW.DRAW();
   DATA_WINDOW.CLEAR();
   DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFfracType);
   DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFdetail);
   DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFxStart);
   DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFxEnd);
   DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFxPivot);
   DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFyStart);
   DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFyEnd);
   DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFyPivot);
   // Draw control buttons.
   TEXT_8_X_13_WHITE( 10, winHeight-20, "Commands");
   MAKE_BUTTON.DRAW_HILITE(1.0, 1.0, 1.0);
   LOAD_BUTTON.DRAW();
   SAVE_BUTTON.DRAW();
   EXPT_BUTTON.DRAW();
   PALR_BUTTON.DRAW();
   PALG_BUTTON.DRAW();
   PALB_BUTTON.DRAW();
   TYPE_BUTTON.DRAW();
   DETP_BUTTON.DRAW();
   DETM_BUTTON.DRAW();
   ZOMI_BUTTON.DRAW();
   ZOMO_BUTTON.DRAW();
   PIVT_BUTTON.DRAW();
   REST_BUTTON.DRAW();
   EXIT_BUTTON.DRAW_HILITE(1.0, 0.1, 0.1);
   UP___BUTTON.HIDE();
   DOWN_BUTTON.HIDE();
   CNCL_BUTTON.HIDE();
   ACPT_BUTTON.HIDE();
}

/****************************************/
/* Button display control.              */
/****************************************/
void HIDE_LOAD_BUTTONS(void) {
   UP___BUTTON.HIDE();
   DOWN_BUTTON.HIDE();
   CNCL_BUTTON.HIDE();
}

void SHOW_LOAD_BUTTONS(void) {
   UP___BUTTON.SHOW();
   DOWN_BUTTON.SHOW();
   CNCL_BUTTON.SHOW();
}

void HIDE_SAVE_BUTTONS(void) {
   ACPT_BUTTON.HIDE();
   CNCL_BUTTON.HIDE();
}

void SHOW_SAVE_BUTTONS(void) {
   ACPT_BUTTON.SHOW();
   CNCL_BUTTON.SHOW();
}

void HIDE_COMMAND_BUTTONS(void) {
   MAKE_BUTTON.HIDE();
   LOAD_BUTTON.HIDE();
   SAVE_BUTTON.HIDE();
   EXPT_BUTTON.HIDE();
   PALR_BUTTON.DRAW();
   PALG_BUTTON.DRAW();
   PALB_BUTTON.DRAW();
   TYPE_BUTTON.HIDE();
   DETP_BUTTON.HIDE();
   DETM_BUTTON.HIDE();
   ZOMI_BUTTON.HIDE();
   ZOMO_BUTTON.HIDE();
   PIVT_BUTTON.HIDE();
   REST_BUTTON.HIDE();
}

void SHOW_COMMAND_BUTTONS(void) {
   MAKE_BUTTON.DRAW_HILITE(1.0, 1.0, 1.0);
   LOAD_BUTTON.SHOW();
   SAVE_BUTTON.SHOW();
   EXPT_BUTTON.SHOW();
   PALR_BUTTON.DRAW();
   PALG_BUTTON.DRAW();
   PALB_BUTTON.DRAW();
   TYPE_BUTTON.SHOW();
   DETP_BUTTON.SHOW();
   DETM_BUTTON.SHOW();
   ZOMI_BUTTON.SHOW();
   ZOMO_BUTTON.SHOW();
   PIVT_BUTTON.SHOW();
   REST_BUTTON.SHOW();
}

/****************************************/
/* Display whole fractal.               */
/****************************************/
void DISPLAY_FRACTAL(WINDOW window) {
int    x, y;
GLint  w = (window.xs + 2);
GLint  h = (window.ye - 3);
uchar  *ptr;
   glDrawBuffer(GL_BACK);
   glPointSize(2);
   for (y = 0; y < fracHeight; y++) {
       ptr = fractal[y];
       for (x = 0; x < fracWidth; x++) {
           glColor3ub(PaletteR[*ptr], PaletteG[*ptr], PaletteB[*ptr]);
           glBegin(GL_POINTS);
             glVertex2i(w + x, h - y);
           glEnd();
           ptr++;
       }
   }
   glPointSize(1);
   glutSwapBuffers();
}

/****************************************/
/* Display row.                         */
/****************************************/
void DISPLAY_FRACTAL_ROW(int row, WINDOW window) {
int    x;
GLint  w = (window.xs + 2);
GLint  h = (window.ye - 3);
uchar  *ptr;
   ptr = fractal[row];
   for (x = 0; x < fracWidth; x++) {
       glColor3ub(PaletteR[*ptr], PaletteG[*ptr], PaletteB[*ptr]);
       glBegin(GL_POINTS);
         glVertex2i(w + x, h - row);
       glEnd();
       ptr++;
   }
}

/****************************************/
/* Display fractal data.                */
/****************************************/
void DISPLAY_FRACTAL_DATA(WINDOW window, int tag) {
GLint  l = window.xs + 5;
GLint  x = window.xs + 73;
GLint  y = window.ye - 20;
char   text[80];
   glDrawBuffer(GL_BACK);
   glPolygonMode(GL_FRONT, GL_FILL);
   glColor3f(0.0, 0.0, 0.0);
   if (tag == DEFfracType) {
      glBegin(GL_QUADS);
        glVertex2i(window.xs, y-4);
        glVertex2i(window.xe, y-4);
        glVertex2i(window.xe, y+12);
        glVertex2i(window.xs, y+12);
      glEnd();
      TEXT_8_X_13_WHITE(l, y, "Type---- ");
      TEXT_8_X_13_WHITE(x+8, y, fracType);
   } else if (tag == DEFdetail) {
      y -= 20;
      glBegin(GL_QUADS);
        glVertex2i(window.xs, y-4);
        glVertex2i(window.xe, y-4);
        glVertex2i(window.xe, y+12);
        glVertex2i(window.xs, y+12);
      glEnd();
      TEXT_8_X_13_WHITE(l, y, "Detail--    ");
      sprintf(text, " %d", detail);
      TEXT_8_X_13_WHITE(x, y, text);
   } else if (tag == DEFxStart) {
      y -= 40;
      glBegin(GL_QUADS);
        glVertex2i(window.xs, y-4);
        glVertex2i(window.xe, y-4);
        glVertex2i(window.xe, y+12);
        glVertex2i(window.xs, y+12);
      glEnd();
      TEXT_8_X_13_WHITE(l, y, "x Start-");
      sprintf(text, "%11.8lf", xStart);
      TEXT_8_X_13_WHITE(x, y, text);
   } else if (tag == DEFxEnd) {
      y -= 55;
      glBegin(GL_QUADS);
        glVertex2i(window.xs, y-4);
        glVertex2i(window.xe, y-4);
        glVertex2i(window.xe, y+12);
        glVertex2i(window.xs, y+12);
      glEnd();
      TEXT_8_X_13_WHITE(l, y, "x End---");
      sprintf(text, "%11.8lf", xEnd);
      TEXT_8_X_13_WHITE(x, y, text);
   } else if (tag == DEFxPivot) {
      y -= 70;
      glBegin(GL_QUADS);
        glVertex2i(window.xs, y-4);
        glVertex2i(window.xe, y-4);
        glVertex2i(window.xe, y+12);
        glVertex2i(window.xs, y+12);
      glEnd();
      TEXT_8_X_13_WHITE(l, y, "x Pivot-");
      sprintf(text, "%11.8lf", xPivot);
      TEXT_8_X_13_WHITE(x, y, text);
   } else if (tag == DEFyStart) {
      y -= 90;
      glBegin(GL_QUADS);
        glVertex2i(window.xs, y-4);
        glVertex2i(window.xe, y-4);
        glVertex2i(window.xe, y+12);
        glVertex2i(window.xs, y+12);
      glEnd();
      TEXT_8_X_13_WHITE(l, y, "y Start-");
      sprintf(text, "%11.8lf", yStart);
      TEXT_8_X_13_WHITE(x, y, text);
   } else if (tag == DEFyEnd) {
      y -= 105;
      glBegin(GL_QUADS);
        glVertex2i(window.xs, y-4);
        glVertex2i(window.xe, y-4);
        glVertex2i(window.xe, y+12);
        glVertex2i(window.xs, y+12);
      glEnd();
      TEXT_8_X_13_WHITE(l, y, "y End---");
      sprintf(text, "%11.8lf", yEnd);
      TEXT_8_X_13_WHITE(x, y, text);
   } else if (tag == DEFyPivot) {
      y -= 125;
      glBegin(GL_QUADS);
        glVertex2i(window.xs, y-4);
        glVertex2i(window.xe, y-4);
        glVertex2i(window.xe, y+12);
        glVertex2i(window.xs, y+12);
      glEnd();
      TEXT_8_X_13_WHITE(l, y, "y Pivot-");
      sprintf(text, "%11.8lf", yPivot);
      TEXT_8_X_13_WHITE(x, y, text);
   }
   glutSwapBuffers();
}

/***********************************************************/
/* PALETTE ROUTINES                                        */
/***********************************************************/
/****************************************/
/* Generate pallete.                    */
/****************************************/
void DISPLAY_PALLETE_CREATE(void) {
int    i;
double c1, c2, c3;
   for (i = 0; i < 255; i++) {
       palette[LIN][i]  = (uchar)i;
       palette[ILIN][i] = (uchar)(255 - i);
       palette[SQRT][i] = (uchar)floor(sqrt((double)i) * 255.0 / 15.9374);
       palette[LOG10][i]= (uchar)floor(log10(i) * 255.0 / 2.4049);
       palette[SIN][i]  = (uchar)floor(sin((double)i/81.0) * 255.0);
       palette[COS][i]  = (uchar)floor(1.0 + cos((double)i/81.0) * 127.0);
       palette[SIN2][i] = (uchar)floor(1.0 + sin((double)i/40.0) * 127.0);
       palette[COS2][i] = (uchar)floor(1.0 + cos((double)i/40.0) * 127.0);
       palette[RNDM][i] = (uchar)floor(rand() % 255);
   }
   for (i = 1; i < 254; i++) {
       c1 = palette[RNDM][i-1];       
       c2 = palette[RNDM][i];       
       c3 = palette[RNDM][i+1];       
       palette[RNDM][i] = (uchar)floor((c1+c2+c3) / 3.0);
   }
}

/****************************************/
/* Apply pallete.                       */
/****************************************/
void DISPLAY_PALLETE_APPLY(void) {
int    i;
   PaletteR[255] = 255;
   PaletteG[255] = 255;
   PaletteB[255] = 255;
   for (i = 0; i < 255; i++) {
       PaletteR[i] = palette[paletteR][i];
       PaletteG[i] = palette[paletteG][i];
       PaletteB[i] = palette[paletteB][i];
   }
}

/***********************************************************/
/* FILE I/O ROUTINES                                       */
/***********************************************************/
/****************************************/
/* File scan.                     	*/
/****************************************/
int  SCAN_FOR_IMAGES(const char Extension[4]) {
int    i;
char   sysCmd[80];
FILE   *disk;
   strcpy(sysCmd, "ls *");
   strcat(sysCmd, Extension);
   strcat(sysCmd, " > fileList");
   system(sysCmd);
   if ((disk = fopen("fileList", "r")) != NULL) {
      for (i = 0; i < maxNames; i++) {
          fscanf(disk, "%s", fileNames[i]);
          if (feof(disk)) break;
      }
   }
   fclose(disk);
   system("rm fileList");
   if (i == 0) return 0;
          else return --i;
}

/****************************************/
/* File scan display.                   */
/****************************************/
void SCAN_FILE_DISPLAY(int start, int fractals) {
int    i, j = start;
   FILE_WINDOW.CLEAR();
   glColor3f(1.0, 1.0, 1.0);
   for (i = 15; i <= 300; i+=20) {
       TEXT_8_X_13_WHITE(FILE_WINDOW.xs+5, FILE_WINDOW.ye-i, fileNames[j]);
       if (j == fractals) break;
       j++;
   }
   glutSwapBuffers();
}

/****************************************/
/* Load fractal.                     	*/
/****************************************/
int  LOAD_FRACTAL(char fileName[80]) {
int    x, y, k = 0;
int    pmax;
char   ch, type[3];
FILE   *disk;
uchar  *ptr;
   disk = fopen(fileName, "rb");
   // Fractal data.
   fscanf(disk, "%s",  fracType);
   if (strcmp(fracType, "FRACRAM")) return 0;
   fscanf(disk, "%s",  fracType);
   if (!strcmp(fracType, "Mandelbrot"))     fType = 0;
   else if (!strcmp(fracType, "Julia"))     fType = 1;
   else if (!strcmp(fracType, "M-Dragon"))  fType = 2;
   else if (!strcmp(fracType, "J-Dragon"))  fType = 3;
   else if (!strcmp(fracType, "M-Phoenix")) fType = 4;
   else if (!strcmp(fracType, "J-Phoenix")) fType = 5;
   else if (!strcmp(fracType, "M-Cosine"))  fType = 6;
   else if (!strcmp(fracType, "J-Cosine"))  fType = 7;
   else if (!strcmp(fracType, "M-Cosh"))    fType = 8;
   else if (!strcmp(fracType, "J-Cosh"))    fType = 9;
   fscanf(disk, "%d",  &detail);
   fscanf(disk, "%lf", &xStart);
   fscanf(disk, "%lf", &xEnd);
   fscanf(disk, "%lf", &xPivot);
   fscanf(disk, "%lf", &yStart);
   fscanf(disk, "%lf", &yEnd);
   fscanf(disk, "%lf", &yPivot);
   // Fractal image.
   fscanf(disk, "%s", type);
   fread(&ch, 1, 1, disk); 
   fscanf(disk, "%d", &fracWidth);
   fscanf(disk, "%d", &fracHeight);
   fscanf(disk, "%d", &pmax);
   fread(&ch, 1, 1, disk);
   for (y = 0; y < fracHeight; y++) {
       ptr = fractal[y];
       for (x = 0; x < fracWidth; x++) {
           *ptr = getc(disk);
           ptr++;
       }
   }
   fclose(disk);
   return 1;
}

/****************************************/
/* Save fractal.                     	*/
/****************************************/
int  SAVE_FRACTAL(char fileName[80]) {
int    x, y;
uchar  *ptr;
FILE   *disk;
   disk = fopen(fileName, "wb");
   // Fractal data.
   fprintf(disk, "FRACRAM\n");
   fprintf(disk, "%s\n",  fracType);
   fprintf(disk, "%d\n",  detail);
   fprintf(disk, "%lf\n", xStart);
   fprintf(disk, "%lf\n", xEnd);
   fprintf(disk, "%lf\n", xPivot);
   fprintf(disk, "%lf\n", yStart);
   fprintf(disk, "%lf\n", yEnd);
   fprintf(disk, "%lf\n", yPivot);
   // Fractal image (as PPM file).
   fprintf(disk, "P5\n");
   fprintf(disk, "%d %d\n", fracWidth, fracHeight);
   fprintf(disk, "255\n");
   for (y = 0; y < fracHeight; y++) {
       ptr = fractal[y];
       for (x = 0; x < fracWidth; x++) {
           fprintf(disk, "%c", *ptr);
           ptr++;
       }
   } 
   fclose(disk);
}

/****************************************/
/* Export fractal.                     	*/
/****************************************/
void EXPORT_FRACTAL(char fileName[80]) {
int    x, y;
uchar  *ptr;
FILE   *disk;
char   tempname[80];
char   command[150];
   disk = fopen("temp.ppm", "wb");
   // Fractal image (as PPM file).
   fprintf(disk, "P5\n");
   fprintf(disk, "%d %d\n", fracWidth, fracHeight);
   fprintf(disk, "255\n");
   for (y = 0; y < fracHeight; y++) {
       ptr = fractal[y];
       for (x = 0; x < fracWidth; x++) {
           fprintf(disk, "%c", *ptr);
           ptr++;
       }
   } 
   fclose(disk);
   // Convert ppm file to png file.
   strcpy(command, "convert ");
   strcat(command, "temp.ppm");
   strcat(command, " " );
   strcat(command, fileName);
   system(command);
}


/***********************************************************/
/* FRACTAL GENERATION ROUTINES                             */
/***********************************************************/
/****************************************/
/* Mandelbrot.                          */
/****************************************/
void MANDELBROT_CORE(int y, uchar *ptr) {
int    x, color;
double xx, yy, xs, ys, xy, cx, cy;
   cy  = yStart + (y * yScale);
   for (x = 0; x < fracWidth; x++) {
       cx    = xStart + (x * xScale);
       xx    = 0.0;
       yy    = 0.0;
       color = -1;
       do {
          xs = (xx * xx);
          ys = (yy * yy);
          xy = (xx * yy);
          xx = (xs - ys + cx);
          yy = (xy + xy + cy);
          color++;
       } while ((color < detail) && ((xs + ys) <= 4.0));
       *ptr = ((color * 255) / detail);
       ptr++;
   }
   return;
}

void MANDELBROT(WINDOW window) {
int y;
uchar *ptr;
   for (y = 0; y < fracHeight; y++) {
       MANDELBROT_CORE(y, fractal[y]);
       DISPLAY_FRACTAL_ROW(y, window);
   } 
}

void *MANDELBROT(void *arg) {
int exitCode = 1;
uchar *ptr;
struct thread_data *params;
   params = ((thread_data*)arg);
   MANDELBROT_CORE(params->row, fractal[params->row]);
   params->status = 1;
   pthread_exit(&exitCode);
}

/****************************************/
/* Julia.                               */
/****************************************/
uchar *JULIA_CORE(int y, uchar *ptr) {
int    x, color;
double xx, yy, xs, ys, xy, cx, cy;
   for (x = 0; x < fracWidth; x++) {
       yy    = yStart + (y * yScale);
       xx    = xStart + (x * xScale);
       color = -1;
       do {
          xs = (xx * xx);
          ys = (yy * yy);
          xy = (xx * yy);
          xx = (xs - ys + xPivot);
          yy = (xy + xy + yPivot);
          color++;
       } while ((color < detail) && ((xs + ys) <= 4.0));
       *ptr = ((color * 255) / detail);
       ptr++;
   }
   return ptr;
}

void JULIA(WINDOW window) {
int y;
uchar *ptr;
   for (y = 0; y < fracHeight; y++) {
       ptr = JULIA_CORE(y, fractal[y]);
       DISPLAY_FRACTAL_ROW(y, window);
   } 
}

void *JULIA(void *arg) {
int exitCode = 1;
uchar *ptr;
struct thread_data *params;
   params = ((thread_data*)arg);
   ptr = JULIA_CORE(params->row, fractal[params->row]);
   params->status = 1;
   pthread_exit(&exitCode);
}

/****************************************/
/* Mandelbrot Self Squared Dragon.      */
/****************************************/
uchar *M_DRAGON_CORE(int y, uchar *ptr) {
int    x, color;
double xx, yy, xs, ys, xt, yt, cx, cy;
   cy  = yStart + (y * yScale);
   for (x = 0; x < fracWidth; x++) {
       cx    = xStart + (x * xScale);
       xx    = 0.5;
       yy    = 0.0;
       color = -1;
       do {
          xs = (xx * xx);
          ys = (yy * yy);
          xt = (ys - xs + xx);
          yt = ((xx * yy * 2) - yy);
          xx = ((cx * xt) + (cy * yt));
          yy = ((cy * xt) - (cx * yt));
          color++;
       } while ((color < detail) && ((xs + ys) <= 4.0));
       *ptr = ((color * 255) / detail);
       ptr++;
   }
   return ptr;
}

void M_DRAGON(WINDOW window) {
int y;
uchar *ptr;
   for (y = 0; y < fracHeight; y++) {
       ptr = M_DRAGON_CORE(y, fractal[y]);
       DISPLAY_FRACTAL_ROW(y, window);
   } 
}

void *M_DRAGON(void *arg) {
int exitCode = 1;
uchar *ptr;
struct thread_data *params;
   params = ((thread_data*)arg);
   ptr = M_DRAGON_CORE(params->row, fractal[params->row]);
   params->status = 1;
   pthread_exit(&exitCode);
}

/****************************************/
/* Julia Self Squared Dragon.           */
/****************************************/
uchar *J_DRAGON_CORE(int y, uchar *ptr) {
int    x, color;
double xx, yy, xs, ys, xy, cx, cy, t1, t2;
   cy  = yStart + (y * yScale);
   for (x = 0; x < fracWidth; x++) {
       cx    = xStart + (x * xScale);
       xx    = cx;
       yy    = cy;
       color = -1;
       do {
          xs = (xx * xx);
          ys = (yy * yy);
          t1 = ((ys - xs) + xx);
          t2 = ((xx * yy * 2) - yy);
          xx = ((xPivot * t1) + (yPivot * t2));
          yy = ((yPivot * t1) - (xPivot * t2));
          color++;
       } while ((color < detail) && ((xs + ys) <= 4.0));
       *ptr = ((color * 255) / detail);
       ptr++;

   }
   return ptr;
}

void J_DRAGON(WINDOW window) {
int y;
uchar *ptr;
   for (y = 0; y < fracHeight; y++) {
       ptr = JULIA_CORE(y, fractal[y]);
       DISPLAY_FRACTAL_ROW(y, window);
   } 
}

void *J_DRAGON(void *arg) {
int exitCode = 1;
uchar *ptr;
struct thread_data *params;
   params = ((thread_data*)arg);
   ptr = JULIA_CORE(params->row, fractal[params->row]);
   params->status = 1;
   pthread_exit(&exitCode);
}

/****************************************/
/* Mandelbrot Phoenix.                  */
/****************************************/
uchar *M_PHOENIX_CORE(int y, uchar *ptr) {
int    x, color;
double xx, yy, xi, yi, xs, ys, xt, yt, cx, cy;
   cy  = yStart + (y * yScale);
   for (x = 0; x < fracWidth; x++) {
       cx    = xStart + (x * xScale);
       xx    = 0.0;
       yy    = 0.0;
       xi    = 0.0;
       yi    = 0.0;
       color = -1;
       do {
          xs = (xx * xx);
          ys = (yy * yy);
          xt = ((xs - ys + cx) + (cy * xi));
          yt = ((xx * yy * 2)  + (cy * yi));
          xi = xx;
          xx = xt;
          yi = yy;
          yy = yt;
          color++;
       } while ((color < detail) && ((xs + ys) <= 4.0));
       *ptr = ((color * 255) / detail);
       ptr++;
   }
   return ptr;
}

void M_PHOENIX(WINDOW window) {
int y;
uchar *ptr;
   for (y = 0; y < fracHeight; y++) {
       ptr = M_PHOENIX_CORE(y, fractal[y]);
       DISPLAY_FRACTAL_ROW(y, window);
   } 
}

void *M_PHOENIX(void *arg) {
int exitCode = 1;
uchar *ptr;
struct thread_data *params;
   params = ((thread_data*)arg);
   ptr = M_PHOENIX_CORE(params->row, fractal[params->row]);
   params->status = 1;
   pthread_exit(&exitCode);
}

/****************************************/
/* Julia Phoenix.                       */
/****************************************/
uchar *J_PHOENIX_CORE(int y, uchar *ptr) {
int    x, color;
double xx, yy, xi, yi, xs, ys, xy, cx, cy, yt;
   cy  = yStart + (y * yScale);
   for (x = 0; x < fracWidth; x++) {
       cx    = xStart + (x * xScale);
       xx    = cx;
       yy    = cy;
       xi    = 0.0;
       yi    = 0.0;
       color = -1;
       do {
          xs = (xx * xx);
          ys = (yy * yy);
          xi = (xs - ys + xPivot + (yPivot * xi));
          yt = (xx * yy * 2 + (yPivot * yi));
          xx = xi;
          yi = yy;
          yy = yt;
          color++;
       } while ((color < detail) && ((xs + ys) <= 4.0));
       *ptr = ((color * 255) / detail);
       ptr++;

   }
   return ptr;
}

void J_PHOENIX(WINDOW window) {
int y;
uchar *ptr;
   for (y = 0; y < fracHeight; y++) {
       ptr = J_PHOENIX_CORE(y, fractal[y]);
       DISPLAY_FRACTAL_ROW(y, window);
   } 
}

void *J_PHOENIX(void *arg) {
int exitCode = 1;
uchar *ptr;
struct thread_data *params;
   params = ((thread_data*)arg);
   ptr = J_PHOENIX_CORE(params->row, fractal[params->row]);
   params->status = 1;
   pthread_exit(&exitCode);
}

/****************************************/
/* Mandelbrot Cosine.                   */
/****************************************/
uchar *M_COSINE_CORE(int y, uchar *ptr) {
int    x, color;
double xx, yy, xs, ys, co, si, cx, cy;
   cy  = yStart + (y * yScale);
   for (x = 0; x < fracWidth; x++) {
       cx    = xStart + (x * xScale);
       xx    = 0.0;
       yy    = 0.0;
       color = -1;
       do {
          xs = (xx * xx);
          ys = (yy * yy);
          co = (cos(xx) * cosh(yy));
          si = (sin(xx) * sinh(yy));
          xx = ((co * co) - (si * si) + cx);
          yy = ((-2 * co * si) + cy);
          color++;
       } while ((color < detail) && ((xs + ys) <= 4.0));
       *ptr = ((color * 255) / detail);
       ptr++;
   }
   return ptr;
}

void M_COSINE(WINDOW window) {
int y;
uchar *ptr;
   for (y = 0; y < fracHeight; y++) {
       ptr = M_COSINE_CORE(y, fractal[y]);
       DISPLAY_FRACTAL_ROW(y, window);
   } 
}

void *M_COSINE(void *arg) {
int exitCode = 1;
uchar *ptr;
struct thread_data *params;
   params = ((thread_data*)arg);
   ptr = M_COSINE_CORE(params->row, fractal[params->row]);
   params->status = 1;
   pthread_exit(&exitCode);
}

/****************************************/
/* Julia Cosine.                        */
/****************************************/
uchar *J_COSINE_CORE(int y, uchar *ptr) {
int    x, color;
double xx, yy, xs, ys, cx, cy, co, si;
   cy  = yStart + (y * yScale);
   for (x = 0; x < fracWidth; x++) {
       cx    = xStart + (x * xScale);
       xx    = cx;
       yy    = cy;
       color = -1;
       do {
          xs = (xx * xx);
          ys = (yy * yy);
          co = (cos(xx) * cosh(yy));
          si = (sin(xx) * sinh(yy));
          xx = (xPivot + (co * co) - (si * si));
          yy = (yPivot - (2 * co * si));
          color++;
       } while ((color < detail) && ((xs + ys) <= 4.0));
       *ptr = ((color * 255) / detail);
       ptr++;

   }
   return ptr;
}

void J_COSINE(WINDOW window) {
int y;
uchar *ptr;
   for (y = 0; y < fracHeight; y++) {
       ptr = J_COSINE_CORE(y, fractal[y]);
       DISPLAY_FRACTAL_ROW(y, window);
   } 
}

void *J_COSINE(void *arg) {
int exitCode = 1;
uchar *ptr;
struct thread_data *params;
   params = ((thread_data*)arg);
   ptr = J_COSINE_CORE(params->row, fractal[params->row]);
   params->status = 1;
   pthread_exit(&exitCode);
}

/****************************************/
/* Mandelbrot Hyperbolic Cosine.        */
/****************************************/
uchar *M_COSH_CORE(int y, uchar *ptr) {
int    x, color;
double xx, yy, xs, ys, co, si, cx, cy;
   cy  = yStart + (y * yScale);
   for (x = 0; x < fracWidth; x++) {
       cx    = xStart + (x * xScale);
       xx    = 0.0;
       yy    = 0.0;
       color = -1;
       do {
          xs = (xx * xx);
          ys = (yy * yy);
          co = (cos(yy) * cosh(xx));
          si = (sin(yy) * sinh(xx));
          xx = (cx + (co * co) - (si * si));
          yy = (cy - (2 * co * si));
          color++;
       } while ((color < detail) && ((xs + ys) <= 4.0));
       *ptr = ((color * 255) / detail);
       ptr++;
   }
   return ptr;
}

void M_COSH(WINDOW window) {
int y;
uchar *ptr;
   for (y = 0; y < fracHeight; y++) {
       ptr = M_COSH_CORE(y, fractal[y]);
       DISPLAY_FRACTAL_ROW(y, window);
   } 
}

void *M_COSH(void *arg) {
int exitCode = 1;
uchar *ptr;
struct thread_data *params;
   params = ((thread_data*)arg);
   ptr = M_COSH_CORE(params->row, fractal[params->row]);
   params->status = 1;
   pthread_exit(&exitCode);
}

/****************************************/
/* Julia Hyperbolic Cosine.             */
/****************************************/
uchar *J_COSH_CORE(int y, uchar *ptr) {
int    x, color;
double xx, yy, xs, ys, cx, cy, co, si;
   cy  = yStart + (y * yScale);
   for (x = 0; x < fracWidth; x++) {
       cx    = xStart + (x * xScale);
       xx    = cx;
       yy    = cy;
       color = -1;
       do {
          xs = (xx * xx);
          ys = (yy * yy);
          co = (cos(yy) * cosh(xx));
          si = (sin(yy) * sinh(xx));
          xx = (xPivot + (co * co) - (si * si));
          yy = (yPivot - (2 * co * si));
          color++;
       } while ((color < detail) && ((xs + ys) <= 4.0));
       *ptr = ((color * 255) / detail);
       ptr++;

   }
   return ptr;
}

void J_COSH(WINDOW window) {
int y;
uchar *ptr;
   for (y = 0; y < fracHeight; y++) {
       ptr = J_COSH_CORE(y, fractal[y]);
       DISPLAY_FRACTAL_ROW(y, window);
   } 
}

void *J_COSH(void *arg) {
int exitCode = 1;
uchar *ptr;
struct thread_data *params;
   params = ((thread_data*)arg);
   ptr = J_COSH_CORE(params->row, fractal[params->row]);
   params->status = 1;
   pthread_exit(&exitCode);
}

/***********************************************************/
/* THREADED FRACTAL GENERATION ROUTINES                    */
/***********************************************************/
/****************************************/
/* Generate fractal using N threads on  */
/* a single machine.                    */
/****************************************/
void THREAD_FRACTAL_GENERATE(WINDOW window) {
int    x, y, t;
int    row, retn, send;
pthread_t fractalRow[threads];
struct thread_data bounds[threads];
   glDrawBuffer(GL_FRONT_AND_BACK);
   glPointSize(2);
    // Divide into rows and send each to a thread.
   row = 0;
   for (t = 0; t < threads; t++) {
       bounds[t].row    = row;
       bounds[t].status = 0;
       if (fType == 0) pthread_create(&fractalRow[t], NULL, MANDELBROT, (void *)&bounds[t]); 
       else if (fType == 1) pthread_create(&fractalRow[t], NULL, JULIA, (void *)&bounds[t]); 
       else if (fType == 2) pthread_create(&fractalRow[t], NULL, M_DRAGON, (void *)&bounds[t]); 
       else if (fType == 3) pthread_create(&fractalRow[t], NULL, J_DRAGON, (void *)&bounds[t]); 
       else if (fType == 4) pthread_create(&fractalRow[t], NULL, M_PHOENIX, (void *)&bounds[t]); 
       else if (fType == 5) pthread_create(&fractalRow[t], NULL, J_PHOENIX, (void *)&bounds[t]); 
       else if (fType == 6) pthread_create(&fractalRow[t], NULL, M_COSINE, (void *)&bounds[t]); 
       else if (fType == 7) pthread_create(&fractalRow[t], NULL, J_COSINE, (void *)&bounds[t]); 
       else if (fType == 8) pthread_create(&fractalRow[t], NULL, M_COSH, (void *)&bounds[t]); 
       else if (fType == 9) pthread_create(&fractalRow[t], NULL, J_COSH, (void *)&bounds[t]); 
       row++;
   }
   // Wait for any thread to complete and resend new data.
   t    = 0;
   retn = 0;
   send = 1;
   while (1) {
      if (bounds[t].status == 1) {
         bounds[t].status = 0;
         pthread_detach(fractalRow[t]);
         retn++;
         DISPLAY_FRACTAL_ROW(bounds[t].row, window);
         // Keep sending data?
         if (send == 1) {
            bounds[t].row = row;
            if (fType == 0) pthread_create(&fractalRow[t], NULL, MANDELBROT, (void *)&bounds[t]); 
            else if (fType == 1) pthread_create(&fractalRow[t], NULL, JULIA, (void *)&bounds[t]); 
            else if (fType == 2) pthread_create(&fractalRow[t], NULL, M_DRAGON, (void *)&bounds[t]); 
            else if (fType == 3) pthread_create(&fractalRow[t], NULL, J_DRAGON, (void *)&bounds[t]); 
            else if (fType == 4) pthread_create(&fractalRow[t], NULL, M_PHOENIX, (void *)&bounds[t]); 
            else if (fType == 5) pthread_create(&fractalRow[t], NULL, J_PHOENIX, (void *)&bounds[t]);
            else if (fType == 6) pthread_create(&fractalRow[t], NULL, M_COSINE, (void *)&bounds[t]);  
            else if (fType == 7) pthread_create(&fractalRow[t], NULL, J_COSINE, (void *)&bounds[t]); 
            else if (fType == 8) pthread_create(&fractalRow[t], NULL, M_COSH, (void *)&bounds[t]); 
            else if (fType == 9) pthread_create(&fractalRow[t], NULL, J_COSH, (void *)&bounds[t]); 
            row++;
         }
         if (row  == fracHeight) send = 0;
         if (retn == fracHeight) break;
      }
      t++;
      if (t == threads) t = 0;
   }
   glPointSize(1);
}


/***********************************************************/
/* CALLBACK FUNTIONS.                                      */
/***********************************************************/
/****************************************/
/* Mouse control.                       */
/****************************************/
void myMouse(int button, int state, int x, int y) {
int  filePicked;
char fileName[80];
time_t start, stop;
   y = (winHeight - y);
   if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN) {
      glFinish();
      /**********************************/
      /* Generate fractal.              */
      /**********************************/
      if (MAKE_BUTTON.SELECTED(x, y)) { 
         if (lockMenu == 1) return;
         MAKE_BUTTON.CLICK();
         MAKE_BUTTON.DRAW();
         MAIN_WINDOW.CLEAR();
         start = time(NULL);
         glDrawBuffer(GL_FRONT_AND_BACK);
         glPointSize(2);
         xScale = (xEnd - xStart) / (double)fracWidth;
         yScale = (yEnd - yStart) / (double)fracHeight;
         if (processType == 1) {
            if (fType == 0) MANDELBROT(MAIN_WINDOW);
            if (fType == 1) JULIA(MAIN_WINDOW);
            if (fType == 2) M_DRAGON(MAIN_WINDOW);
            if (fType == 3) J_DRAGON(MAIN_WINDOW);
            if (fType == 4) M_PHOENIX(MAIN_WINDOW);
            if (fType == 5) J_PHOENIX(MAIN_WINDOW);
            if (fType == 6) M_COSINE(MAIN_WINDOW);
            if (fType == 7) J_COSINE(MAIN_WINDOW);
            if (fType == 8) M_COSH(MAIN_WINDOW);
            if (fType == 9) J_COSH(MAIN_WINDOW);
         }
         if (processType == 2) THREAD_FRACTAL_GENERATE(MAIN_WINDOW);
         glPointSize(1);
         stop = time(NULL);
         printf("Run time (sec) = %d.\n", (stop - start));
         noFractal = 0;
         return;
      }
      /**********************************/
      /* Load fractal from disk.  	*/
      /**********************************/
      if (LOAD_BUTTON.SELECTED(x, y)) { 
         LOAD_BUTTON.CLICK();
         lockMenu  = 1;
         SHOW_LOAD_BUTTONS();
         typeFlag  = 1;		// 1 -> .frc file.
         scanStart = 0;
         fractals  = SCAN_FOR_IMAGES("frc");
         SCAN_FILE_DISPLAY(scanStart, fractals);
         return;
      }
      /**********************************/
      /* Save fractal to disk.          */
      /**********************************/
      if (SAVE_BUTTON.SELECTED(x, y)) { 
         if (lockMenu == 1 || noFractal == 1) return;
         SAVE_BUTTON.CLICK();
         lockMenu  = 1;
         strcpy(fileNames[0], "");
         SHOW_SAVE_BUTTONS();
         FILE_WINDOW.CLEAR();
         TEXT_8_X_13_WHITE(FILE_WINDOW.xs+5, FILE_WINDOW.ye-15 , "Enter filename:");
         glutSwapBuffers();
         typeFlag  = -1;	// 1 -> .frc file.
         return;
      }
      /**********************************/
      /* Export fractal to PPM file.    */
      /**********************************/
      if (EXPT_BUTTON.SELECTED(x, y)) { 
         if (lockMenu == 1 || noFractal == 1) return;
         EXPT_BUTTON.CLICK();
         lockMenu  = 1;
         strcpy(fileNames[0], "");
         SHOW_SAVE_BUTTONS();
         FILE_WINDOW.CLEAR();
         TEXT_8_X_13_WHITE(FILE_WINDOW.xs+5, FILE_WINDOW.ye-15 , "Enter filename:");
         glutSwapBuffers();
         typeFlag  = -2;	// 2 -> .png file.
         return;
      }
      /**********************************/
      /* Select palette.                */
      /**********************************/
      if (PALR_BUTTON.SELECTED(x, y)) { 
         if (lockMenu == 1 || noFractal == 1) return;
         PALR_BUTTON.CLICK();
         paletteR++;
         if (paletteR > maxPalette) paletteR = 0;
         DISPLAY_PALLETE_APPLY();
         DISPLAY_FRACTAL(MAIN_WINDOW);
         return;
      }
      if (PALG_BUTTON.SELECTED(x, y)) { 
         if (lockMenu == 1 || noFractal == 1) return;
         PALG_BUTTON.CLICK();
         paletteG++;
         if (paletteG > maxPalette) paletteG = 0;
         DISPLAY_PALLETE_APPLY();
         DISPLAY_FRACTAL(MAIN_WINDOW);
         return;
      }
      if (PALB_BUTTON.SELECTED(x, y)) { 
         if (lockMenu == 1 || noFractal == 1) return;
         PALB_BUTTON.CLICK();
         paletteB++;
         if (paletteB > maxPalette) paletteB = 0;
         DISPLAY_PALLETE_APPLY();
         DISPLAY_FRACTAL(MAIN_WINDOW);
         return;
      }

      /**********************************/
      /* Select fractal type.           */
      /**********************************/
      if (TYPE_BUTTON.SELECTED(x, y)) { 
         if (lockMenu == 1) return;
         TYPE_BUTTON.CLICK();
         fType++;
         if (fType > MaxfracType) fType = 0;
         if (fType == 0) strcpy(fracType, "Mandelbrot");   
         else if (fType == 1) strcpy(fracType, "Julia");   
         else if (fType == 2) strcpy(fracType, "M-Dragon"); 
         else if (fType == 3) strcpy(fracType, "J-Dragon");  
         else if (fType == 4) strcpy(fracType, "M-Phoenix");     
         else if (fType == 5) strcpy(fracType, "J-Phoenix");        
         else if (fType == 6) strcpy(fracType, "M-Cosine");        
         else if (fType == 7) strcpy(fracType, "J-Cosine");        
         else if (fType == 8) strcpy(fracType, "M-Cosh");        
         else if (fType == 9) strcpy(fracType, "J-Cosh");        
         DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFfracType);
         MAKE_BUTTON.DRAW_HILITE(1.0, 1.0, 1.0);
         return;
      }
      /**********************************/
      /* Increase fractal detail.       */        
      /**********************************/
      if (DETP_BUTTON.SELECTED(x, y)) { 
         if (lockMenu == 1) return;
         DETP_BUTTON.CLICK();
         detail+=64;
         if (detail > maxDetail) detail = maxDetail;
         DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFdetail);
         MAKE_BUTTON.DRAW_HILITE(1.0, 1.0, 1.0);
         return;
      }
      /**********************************/
      /* Decrease fractal detail.       */        
      /**********************************/
      if (DETM_BUTTON.SELECTED(x, y)) { 
         if (lockMenu == 1) return;
         DETM_BUTTON.CLICK();
         detail-=64;
         if (detail < 64) detail = 64;
         DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFdetail);
         MAKE_BUTTON.DRAW_HILITE(1.0, 1.0, 1.0);
         return;
      }
      /**********************************/
      /* Zoom in.                       */        
      /**********************************/
      if (ZOMI_BUTTON.SELECTED(x, y)) { 
         if (lockMenu == 1 || noFractal == 1) return;
         ZOMI_BUTTON.CLICK();
         lockMenu  = 1;
         HIDE_COMMAND_BUTTONS();
         glutSetCursor(GLUT_CURSOR_CROSSHAIR);
         zoomFlag = 1;
         return;
      }
      /**********************************/
      /* Zoom out.                      */        
      /**********************************/
      if (ZOMO_BUTTON.SELECTED(x, y)) { 
         if (lockMenu == 1 || noFractal == 1) return;
         ZOMO_BUTTON.CLICK();
         xStart    = -5.0;
         xEnd      =  5.0;
         xScale    = 10.0 / (double)fracWidth;
         yStart    = -5.0;
         yEnd      =  5.0;
         yScale    = 10.0 / (double)fracHeight;
         DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFxStart);
         DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFxEnd);
         DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFyStart);
         DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFyEnd);
         MAKE_BUTTON.DRAW_HILITE(1.0, 1.0, 1.0);
         return;
      }
      /**********************************/
      /* Select pivot.                  */        
      /**********************************/
      if (PIVT_BUTTON.SELECTED(x, y)) { 
         if (lockMenu == 1 || noFractal == 1) return;
         PIVT_BUTTON.CLICK();
         lockMenu  = 1;
         HIDE_COMMAND_BUTTONS();
         glutSetCursor(GLUT_CURSOR_CROSSHAIR);
         zoomFlag = -1;
         return;
      }
      /**********************************/
      /* Reset parameters.              */        
      /**********************************/
      if (REST_BUTTON.SELECTED(x, y)) { 
         if (lockMenu == 1) return;
         REST_BUTTON.CLICK();
         paletteR  = 0;
         paletteG  = 0;
         paletteB  = 0;
         detail    = 50;
         xStart    = -5.0;
         xEnd      =  5.0;
         xPivot    =  0.0;
         xScale    = 10.0 / (double)fracWidth;
         yStart    = -5.0;
         yEnd      =  5.0;
         yPivot    =  0.0;
         yScale    = 10.0 / (double)fracHeight;
         strcpy(fracType, "Mandelbrot");   
         DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFfracType);
         DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFdetail);
         DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFxStart);
         DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFxEnd);
         DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFxPivot);
         DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFyStart);
         DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFyEnd);
         DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFyPivot);
         DISPLAY_PALLETE_APPLY();
         MAIN_WINDOW.CLEAR();
         MAKE_BUTTON.DRAW_HILITE(1.0, 1.0, 1.0);
      }

      /**********************************/
      /* Exit.                          */        
      /**********************************/
      if (EXIT_BUTTON.SELECTED(x, y)) { 
         EXIT_BUTTON.CLICK();
         glutDestroyWindow(window);
         exit(0);                   
      }

      /**********************************/
      /* Directory listing control.     */        
      /**********************************/
      if (UP___BUTTON.SELECTED(x, y)) { 
         UP___BUTTON.CLICK();
         if (scanStart <= (fractals-15)) {
            scanStart++;
            SCAN_FILE_DISPLAY(scanStart, fractals);
         }
         return;
      }
      if (DOWN_BUTTON.SELECTED(x, y)) { 
         DOWN_BUTTON.CLICK();
         if (scanStart > 0) {
            scanStart--;
            SCAN_FILE_DISPLAY(scanStart, fractals);
         }
         return;
      }
      if (CNCL_BUTTON.SELECTED(x, y)) { 
         CNCL_BUTTON.CLICK();
         FILE_WINDOW.CLEAR();
         if (typeFlag == 1) {
            HIDE_LOAD_BUTTONS();
            HIDE_SAVE_BUTTONS();
         }
         if (typeFlag == -1) {
            HIDE_SAVE_BUTTONS();
            HIDE_SAVE_BUTTONS();
         }
         lockMenu = 0;
         return;
      }
      if (ACPT_BUTTON.SELECTED(x, y)) { 
         ACPT_BUTTON.CLICK();
         FILE_WINDOW.CLEAR();
         HIDE_SAVE_BUTTONS();
         if (typeFlag == -1) {
            strcat(fileNames[0], ".frc");
            SAVE_FRACTAL(fileNames[0]);
         }
         if (typeFlag == -2) {
            strcat(fileNames[0], ".png");
            EXPORT_FRACTAL(fileNames[0]);
         }
         lockMenu = 0;
         return;
      }
      // Filename selection.
      if (x >= 10 && x <= 170 && typeFlag == 1) {
         filePicked = scanStart + ((785 - y) / 20) - 15;
         if (filePicked < 0 || filePicked > fractals) return;
         strcpy(fileName, fileNames[filePicked]);
         if (LOAD_FRACTAL(fileName) == 1) {
            MAIN_WINDOW.CLEAR();
            DISPLAY_FRACTAL(MAIN_WINDOW);
            DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFfracType);
            DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFdetail);
            DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFxStart);
            DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFxEnd);
            DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFxPivot);
            DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFyStart);
            DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFyEnd);
            DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFyPivot);
         }
         typeFlag = 0;
         FILE_WINDOW.CLEAR();
         HIDE_LOAD_BUTTONS();
         noFractal= 0;
         lockMenu = 0;
         xScale   = (xEnd - xStart) / (double)fracWidth;
         yScale   = (yEnd - yStart) / (double)fracHeight;
         return;
      }
   }

   /**********************************/
   /* Zoom control.                  */        
   /**********************************/
   if (zoomFlag == 1 && button == GLUT_LEFT_BUTTON && state == GLUT_DOWN) {
      glFinish();
      xTemp    = xStart + ((x - MAIN_WINDOW.xs) * xScale);
      yTemp    = yStart + ((MAIN_WINDOW.ye - y) * yScale);
      xs       = x;
      ys       = y;
      zoomFlag = 2;
      glutSetCursor(GLUT_CURSOR_NONE);
   }
   if (zoomFlag == 2 && button == GLUT_LEFT_BUTTON && state == GLUT_UP) {
      glFinish();
      glDrawBuffer(GL_BACK);
      glPolygonMode(GL_FRONT, GL_LINE);
      glColor3ub(128, 128, 128);
      glBegin(GL_QUADS);
        glVertex2i(xs, ys);
        glVertex2i(xs, ye);
        glVertex2i(xe, ye);
        glVertex2i(xe, ys);
      glEnd();
      glutSwapBuffers();
      xEnd     = xStart + ((x - MAIN_WINDOW.xs) * xScale);
      yEnd     = yStart + ((MAIN_WINDOW.ye - y) * yScale);
      xStart   = xTemp;
      yStart   = yTemp;
      zoomFlag = 0;
      DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFxStart);
      DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFxEnd);
      DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFyStart);
      DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFyEnd);
      lockMenu = 0;
      glutSetCursor(GLUT_CURSOR_INFO);
      SHOW_COMMAND_BUTTONS();
   }

   /**********************************/
   /* Pivot control.                 */        
   /**********************************/
   if (zoomFlag == -1 && button == GLUT_LEFT_BUTTON && state == GLUT_DOWN) {
      glFinish();
      glDrawBuffer(GL_BACK);
      glPointSize(10);
      glColor3ub(128, 128, 128);
      glBegin(GL_POINTS);
        glVertex2i(x, y);
      glEnd();
      glPointSize(1);
      glutSwapBuffers();
      xPivot   = xStart + ((x - MAIN_WINDOW.xs) * xScale);
      yPivot   = yStart + ((MAIN_WINDOW.ye - y) * yScale);
      zoomFlag = 0;
      DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFxPivot);
      DISPLAY_FRACTAL_DATA(DATA_WINDOW, DEFyPivot);
      lockMenu = 0;
      glutSetCursor(GLUT_CURSOR_INFO);
      SHOW_COMMAND_BUTTONS();
   }
}

/****************************************/
/* Mouse control.                       */
/****************************************/
void myMovingMouse(int x, int y) {
   if (zoomFlag != 2) return;
   xe = x;
   ye = (winHeight - y); 
   glDrawBuffer(GL_FRONT);
   glPolygonMode(GL_FRONT, GL_LINE);
   glColor3ub(128, 128, 128);
   glBegin(GL_QUADS);
     glVertex2i(xs, ys);
     glVertex2i(xs, ye);
     glVertex2i(xe, ye);
     glVertex2i(xe, ys);
   glEnd();
}

/****************************************/
/* File prompt.                         */
/****************************************/
void myKeyboard(uchar key, int x, int y) {
char txt[2];
   if (typeFlag > -1) return;
   // Clear text area.
   glClear(GL_DEPTH_BUFFER_BIT);                 
   glDrawBuffer(GL_BACK);
   glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
   glColor3f(0.0, 0.0, 0.0);
   glBegin(GL_QUADS);
     glVertex2i(FILE_WINDOW.xs, FILE_WINDOW.ye-17);
     glVertex2i(FILE_WINDOW.xe, FILE_WINDOW.ye-17);
     glVertex2i(FILE_WINDOW.xe, FILE_WINDOW.ye-32);
     glVertex2i(FILE_WINDOW.xs, FILE_WINDOW.ye-32);
   glEnd();
   glutSwapBuffers();
   // Display filename entered.
   if ((int)key == 8) {
      fileNames[0][strlen(fileNames[0])-1] = '\0';
   } 
   if ((int)key > 31) {
      txt[0] = key;
      txt[1] = '\0';
      strcat(fileNames[0], txt);
   }
   TEXT_8_X_13_WHITE(FILE_WINDOW.xs+5, FILE_WINDOW.ye-30, fileNames[0]);
   glutSwapBuffers();
}

/****************************************/
/* Resize.                              */
/****************************************/
void myResize(int width, int height) {
   glutReshapeWindow(winWidth, winHeight);                    
}

/****************************************/
/* Display.                             */
/****************************************/
void myDisplay(void) {
   if (drawFlag == 0) {
      glutSwapBuffers();
   } else {
      DISPLAY_MAIN_PANEL();
      drawFlag = 0;
   }
}

/****************************************/
/* Init.                                */
/****************************************/
void myInit(void) {
   // initialize OpenGL parameters.
   glPointSize(1);
   glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
   glClear(GL_COLOR_BUFFER_BIT); 
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity(); 
   gluOrtho2D(0.0, (GLdouble)winWidth, 0.0, (GLdouble)winHeight);
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();     
   // Initialize prgram parameters.
   strcpy(fracType, "Mandelbrot");   
   srand(1959);
   DISPLAY_PALLETE_CREATE();  
   DISPLAY_PALLETE_APPLY();
}


/***********************************************************/
/* MAIN.                                                   */
/***********************************************************/
/****************************************/
/* Main.                                */
/****************************************/
int main(int argc, char **argv) {
int  nodesPossible;
char text[8];
char title[80];
   strcpy(title, "Frac-R.A.M. (V. ");
   strcat(title, version);
   // Trap HELP option.
   if (argc == 2) { 
      system("clear");
      printf("%s\n", title);
      printf("\n");
      printf("Frac-R.A.M. is designed to support multiple modes of processing:\n");
      printf("\n");
      printf(" 1. fracram - Executes Frac-R.A.M. using a single thread.\n");
      printf("\n");
      printf(" 2. fracram t N - Executes Frac-R.A.M. using a multiple (N) threads.\n");
      printf("    Note that Frac-R.A.M. is limited to 100 threads (N<100).\n");
      printf("\n");
      exit(0);
   }
   // Mode 1 - Trap uni-process option.
   if (argc == 1) {
      processType = 1;
      threads     = 1;
      strcat(title, " - UniProcess");
   }
   // Mode 2 - Trap threaded-process option.
   if (argc == 3 && argv[1][0] == 't') {
      threads     = atoi(argv[2]);
      if (threads == 0) {
         processType = 1;
         threads     = 1;
         strcat(title, " - UniProcess");
      } else {
         processType = 2;
         threads     = atoi(argv[2]);
         if (threads > 99) threads = 99;	// Limit to 100 threads.
         sprintf(text, "%d", (threads+1));
         strcat(title, " - Threaded x ");
         strcat(title, text);
      }
   }
   strcat(title, ") by R. Marsh.");
   
   // Start GUI.
   glutInit(&argc, argv);
   glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);  
   glutInitWindowSize(winWidth, winHeight);
   glutInitWindowPosition(10, 10);
   window = glutCreateWindow(title);
   glutSetCursor(GLUT_CURSOR_INFO);
   glutDisplayFunc(myDisplay);
   glutReshapeFunc(myResize);
   glutMouseFunc(myMouse);
   glutMotionFunc(myMovingMouse);
   glutKeyboardFunc(myKeyboard);
   myInit();
   glutMainLoop();

   return 1;
}
