#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <signal.h>
#include <time.h>
#include <unistd.h>
#include <errno.h>
#include <pthread.h>

#include "libRAM.h"

/***********************************************************/
/* GLOBAL DEFINES                                          */
/***********************************************************/
typedef unsigned char uchar;

#define version      "4.0"

#define DEFfracType  1
#define MaxfracType  9
#define DEFdetail    2
#define DEFxStart    3
#define DEFxEnd      4
#define DEFxPivot    5
#define DEFyStart    6
#define DEFyEnd      7
#define DEFyPivot    8

#define fractWidth  800
#define fractHieght 800
#define maxNames    100
#define maxThreads  100

#define maxPalette  9
#define LIN         0
#define ILIN        1
#define SQRT        2
#define LOG10       3
#define SIN         4
#define SIN2        5
#define COS         6
#define COS2        7
#define RNDM        8
#define maxDetail   765

/***********************************************************/
/* THREAD STRUCTURE                                        */
/***********************************************************/
struct thread_data {
   int row;
   int status;
};

/***********************************************************/
/* PROTOTYPES                                              */
/***********************************************************/
void DISPLAY_MAIN_PANEL(void);
void HIDE_LOAD_BUTTONS(void);
void SHOW_LOAD_BUTTONS(void);
void HIDE_SAVE_BUTTONS(void);
void SHOW_SAVE_BUTTONS(void);
void HIDE_COMMAND_BUTTONS(void);
void SHOW_COMMAND_BUTTONS(void);

void DISPLAY_FRACTAL(WINDOW);
void DISPLAY_FRACTAL_ROW(int, WINDOW);
void DISPLAY_FRACTAL_DATA(WINDOW, int);
void DISPLAY_PALLETE_CREATE(void);
void DISPLAY_PALLETE_APPLY(void);

int  SCAN_FOR_IMAGES(char Extension[4]);
void SCAN_FILE_DISPLAY(int, int);
int  LOAD_FRACTAL(char fileName[80]);
int  SAVE_FRACTAL(char fileName[80]);
void EXPORT_FRACTAL(char fileName[80]);

void MANDELBROT(int);
void GENERATE_FRACTAL_UNIPROCESS(WINDOW);
void *MANDELBROT(void*);
void *JULIA(void*);
void THREAD_FRACTAL_GENERATE(WINDOW);


/***********************************************************/
/* GLOBAL VARIABLES                                        */
/***********************************************************/
// For OpenGL.
GLint  window;
GLint  fracWidth = fractWidth;
GLint  fracHeight= fractHieght;
GLint  winWidth  = fracWidth + 195;
GLint  winHeight = fracHeight+ 20;
int    drawFlag  = 1;

// For Fractals.
int    paletteR  = 0; 
int    paletteG  = 0; 
int    paletteB  = 0; 
uchar  PaletteR[256], PaletteG[256], PaletteB[256], palette[maxPalette][256];
int    fType     = 0;
int    detail    = 64;
double xStart    = -5.0; 
double xEnd      =  5.0;
double xPivot    =  0.0;
double xScale    = 10.0 / (double)fracWidth;
double yStart    = -5.0;
double yEnd      =  5.0;
double yPivot    =  0.0;
double yScale    = 10.0 / (double)fracHeight;
char   fracType[80];
uchar  fractal[fractHieght][fractWidth];  

// For Controls.
int    processType;
int    fractals;
int    noFractal = 1;
int    lockMenu  = 0;
int    typeFlag  = 0;
int    scanStart = 0;
int    zoomFlag  = 0;
int    xs, ys, xe, ye;
double xTemp, yTemp;
char   fileNames[maxNames][80];

// For thread.
int    threads   = 0;
int    thstate[maxThreads][2];


/***********************************************************/
/* OBJECT CONTRUCTION                                      */
/***********************************************************/
// Display panel.
WINDOW MAIN_WINDOW(185,  10, fracWidth+4, fracHeight+4); 
WINDOW FILE_WINDOW( 10, 190, 165, 300); 
WINDOW DATA_WINDOW( 10,  10, 165, 150); 

// Control panel.
BUTTON MAKE_BUTTON( 10, winHeight- 50, "Make");
BUTTON LOAD_BUTTON( 10, winHeight- 75, "Load");
BUTTON SAVE_BUTTON( 10, winHeight-100, "Save");
BUTTON EXPT_BUTTON( 10, winHeight-125, "Export");

BUTTON PALR_BUTTON( 10, winHeight-175, "Red");
BUTTON PALG_BUTTON( 10, winHeight-200, "Green");
BUTTON PALB_BUTTON( 10, winHeight-225, "Blue");

BUTTON TYPE_BUTTON(105, winHeight- 50, "Type");
BUTTON DETP_BUTTON(105, winHeight- 75, "Detail+");
BUTTON DETM_BUTTON(105, winHeight-100, "Detail-");
BUTTON ZOMI_BUTTON(105, winHeight-125, "Zoom In");
BUTTON ZOMO_BUTTON(105, winHeight-150, "Zoom Out");
BUTTON PIVT_BUTTON(105, winHeight-175, "Pivot");
BUTTON REST_BUTTON(105, winHeight-200, "Reset");

BUTTON EXIT_BUTTON(105, winHeight-225, "Exit");

BUTTON UP___BUTTON( 10, winHeight-295, "Up");
BUTTON DOWN_BUTTON( 10, winHeight-320, "Down");
BUTTON CNCL_BUTTON(105, winHeight-320, "Cancel");
BUTTON ACPT_BUTTON(105, winHeight-295, "Accept");

